/**
	$Id: FormatterFactory.java 6268 2013-04-11 06:51:47Z nicklas $

	Copyright (C) 2006 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.formatter;

import java.util.Date;

import net.sf.basedb.clients.web.Base;
import net.sf.basedb.core.Annotation;
import net.sf.basedb.core.Coloring;
import net.sf.basedb.core.ExtendedProperty;
import net.sf.basedb.core.Formula;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.Unit;
import net.sf.basedb.util.ColorGenerator;
import net.sf.basedb.util.formatter.AdaptiveNumberFormatter;
import net.sf.basedb.util.formatter.BooleanFormatter;
import net.sf.basedb.util.formatter.DateFormatter;
import net.sf.basedb.util.formatter.Formatter;
import net.sf.basedb.util.formatter.IntegerFormatter;
import net.sf.basedb.util.formatter.NumberFormatter;
import net.sf.basedb.util.formatter.PrefixSuffixFormatter;
import net.sf.basedb.util.formatter.ToStringFormatter;

/**
	Factory class for creating formatters for various data types. Some formatters
	are affected by user settings on the session control. For example, the number
	of decimals to display by a number formatter or the date format of a date
	formatter.

	@author nicklas
	@version 2.2
	@base.modified $Date: 2013-04-11 08:51:47 +0200 (to, 11 apr 2013) $
*/
public class FormatterFactory
{

	/**
		Get the logged in user's number formatter for float values. The number of 
		decimals to display is configured by the user setting 
		<code>formatter.number.decimals</code>. The default number of decimals is two.

		@param sc The session control of the logged in user
		@return A number formatter
		@see FormatterSettings#getNumDecimals(SessionControl)
		@see FormatterSettings#setNumDecimals(SessionControl, int)
		@see NumberFormatter
	*/
	@SuppressWarnings("unchecked")
	public static Formatter<Number> getNumberFormatter(SessionControl sc)
	{
		Formatter<Number> formatter = (Formatter<Number>)sc.getSessionSetting("formatter.float");
		if (formatter == null)
		{
			formatter = new AdaptiveNumberFormatter(FormatterSettings.getNumDecimals(sc), false);
			sc.setSessionSetting("formatter.float", formatter);
		}
		return formatter;
	}

	/**
		Get the logged in user's number formatter for double values. The number of 
		decimals to display is configured by the user setting 
		<code>formatter.number.decimals</code>. The default number of decimals is two.
	
		@param sc The session control of the logged in user
		@return A number formatter
		@see FormatterSettings#getNumDecimals(SessionControl)
		@see FormatterSettings#setNumDecimals(SessionControl, int)
		@see NumberFormatter
	*/
	@SuppressWarnings("unchecked")
	public static Formatter<Number> getDoubleFormatter(SessionControl sc)
	{
		Formatter<Number> formatter = (Formatter<Number>)sc.getSessionSetting("formatter.double");
		if (formatter == null)
		{
			formatter = new AdaptiveNumberFormatter(FormatterSettings.getNumDecimals(sc), true);
			sc.setSessionSetting("formatter.double", formatter);
		}
		return formatter;
	}
	
	
	/**
		Get the logged in user's formatter for integral numbers. This formatter has no
		configuration options.
		@param sc The session control of the logged in user
		@return A number formatter
		@see IntegerFormatter
	*/
	@SuppressWarnings("unchecked")
	public static Formatter<Number> getIntFormatter(SessionControl sc)
	{
		Formatter<Number> formatter = (Formatter<Number>)sc.getSessionSetting("formatter.integer");
		if (formatter == null)
		{
			formatter = new IntegerFormatter(false);
			sc.setSessionSetting("formatter.integer", formatter);
		}
		return formatter;
	}

	/**
		Get the logged in user's formatter for long numbers. This formatter has no
		configuration options.
		@param sc The session control of the logged in user
		@return A number formatter
		@see IntegerFormatter
	*/
	@SuppressWarnings("unchecked")
	public static Formatter<Number> getLongFormatter(SessionControl sc)
	{
		Formatter<Number> formatter = (Formatter<Number>)sc.getSessionSetting("formatter.long");
		if (formatter == null)
		{
			formatter = new IntegerFormatter(true);
			sc.setSessionSetting("formatter.long", formatter);
		}
		return formatter;
	}

	
	/**
		Get the logged in user's date formatter. How the date is displayed is configured by 
		the user setting <code>formatter.date.formatstring</code>. The default date format
		is <code>yyyy-MM-dd</code>.
	
		@param sc The session control of the logged in user
		@return A date formatter
		@see FormatterSettings#getDateFormat(SessionControl)
		@see FormatterSettings#setDateFormat(SessionControl, String)
		@see DateFormatter
	*/
	@SuppressWarnings("unchecked")
	public static Formatter<Date> getDateFormatter(SessionControl sc)
	{
		Formatter<Date> formatter = (Formatter<Date>)sc.getSessionSetting("formatter.date");
		if (formatter == null)
		{
			formatter = new DateFormatter(FormatterSettings.getDateFormat(sc));
			sc.setSessionSetting("formatter.date", formatter);
		}
		return formatter;
	}

	/**
		Get the logged in user's datetime formatter. How the date and time is displayed is 
		configured by the user setting <code>formatter.datetime.formatstring</code>. The 
		default format is <code>yyyy-MM-dd HH:nn:ss</code>.
	
		@param sc The session control of the logged in user
		@return A date and time formatter
		@see FormatterSettings#getDateTimeFormat(SessionControl)
		@see FormatterSettings#setDateTimeFormat(SessionControl, String)
		@see DateFormatter
	*/
	@SuppressWarnings("unchecked")
	public static Formatter<Date> getDateTimeFormatter(SessionControl sc)
	{
		Formatter<Date> formatter = (Formatter<Date>)sc.getSessionSetting("formatter.datetime");
		if (formatter == null)
		{
			formatter = new DateFormatter(FormatterSettings.getDateTimeFormat(sc));
			sc.setSessionSetting("formatter.datetime", formatter);
		}
		return formatter;
	}
	
	/**
		Get the logged in user's formatter for strings. This formatter has no
		configuration options.
		@param sc The session control of the logged in user
		@return A string formatter
		@see StringFormatter
	*/
	@SuppressWarnings("unchecked")
	public static Formatter<String> getStringFormatter(SessionControl sc)
	{
		Formatter<String> formatter = (Formatter<String>)sc.getSessionSetting("formatter.string");
		if (formatter == null)
		{
			formatter = new StringFormatter();
			sc.setSessionSetting("formatter.string", formatter);
		}
		return formatter;
	}

	/**
		Get the logged in user's formatter for boolean values. This formatter has no
		configuration options.
		@param sc The session control of the logged in user
		@return A boolean formatter
		@see BooleanFormatter
	*/
	@SuppressWarnings("unchecked")
	public static Formatter<Boolean> getBooleanFormatter(SessionControl sc)
	{
		Formatter<Boolean> formatter = (Formatter<Boolean>)sc.getSessionSetting("formatter.boolean");
		if (formatter == null)
		{
			formatter = new BooleanFormatter(BooleanFormatter.Type.TRUE_FALSE);
			sc.setSessionSetting("formatter.boolean", formatter);
		}
		return formatter;
	}

	/**
		Get a formatter suitable for displaying values of the given type.
		If type is <code>FLOAT</code> or <code>DOUBLE</code> a {@link #getNumberFormatter(SessionControl)}
		is returned. If type is <code>LONG</code> or <code>INT</code>
		a {@link #getIntFormatter(SessionControl)} is returned. If type is <code>STRING</code>
		or <code>TEXT</code> a {@link #getStringFormatter(SessionControl)} is returned.
		If type is <code>DATE</code> a {@link #getDateFormatter(SessionControl)} is returned.
		If type is <code>BOOLEAN</code> a {@link #getBooleanFormatter(SessionControl)}
		is returned. Otherwise null is returned.
		
		@param sc The logged in user's session control
		@param type The type to get a formatter for
		@return A formatter or null if no suitable formatter could be found
	*/
	public static Formatter<?> getTypeFormatter(SessionControl sc, Type type)
	{
		if (type == Type.FLOAT)
		{
			return getNumberFormatter(sc);
		}
		else if (type == Type.DOUBLE)
		{
			return getDoubleFormatter(sc);
		}
		else if (type == Type.STRING || type == Type.TEXT)
		{
			return getStringFormatter(sc);
		}
		else if (type == Type.INT)
		{
			return getIntFormatter(sc);
		}
		else if (type == Type.LONG)
		{
			return getLongFormatter(sc);
		}
		else if (type == Type.DATE)
		{
			return getDateFormatter(sc);
		}
		else if (type == Type.TIMESTAMP)
		{
			return getDateTimeFormatter(sc);
		}
		else if (type == Type.BOOLEAN)
		{
			return getBooleanFormatter(sc);
		}
		return null;
	}
	
	/**
		Get a formatter for an extended property. This formatter will check the
		{@link ExtendedProperty#getUrl(Object)} if the property value should be
		linked and create a link if needed. The actual value is formatted by
		creating a formatter based on the type of the extended property using 
		the {@link #getTypeFormatter(SessionControl, Type)} method.
		
		@param sc The logged in user's session control
		@param ep The extended property
		@return A formatter for the extended property
		@see ExtendedPropertyFormatter
	*/
	@SuppressWarnings("unchecked")
	public static Formatter<?> getExtendedPropertyFormatter(SessionControl sc, ExtendedProperty ep)
	{
		return new ExtendedPropertyFormatter(ep, getTypeFormatter(sc, ep.getType()));
	}
	
	/**
		Get a formatter that adds a color box to the display which is a visual representation
		of the value. 	A {@link ColorGenerator} is created based on the {@link Coloring}
		information. The color generator is used to create the color for a specific value.
		The value is formatted by the supplied formatter and is not affected by the color
		formatter. If the {@link Coloring#isUsingColors()} returns false no color formatter
		is created and the orginal formatter is returned. 
		
		@param sc The logged in user's session control
		@param coloring The coloring object that keeps information about how
			to generate colors
		@param valueFormatter A formatter for the actual values
		@return A formatter that adds color to the values
		@see ColorFormatter
	*/
	public static Formatter<Number> getColorFormatter(SessionControl sc, Coloring coloring, Formatter<Number> valueFormatter)
	{
		Formatter<Number> formatter = valueFormatter;
		if (coloring != null && coloring.isUsingColors())
		{
			ColorGenerator colorGen = Base.getColorGenerator(sc, coloring);
			formatter = new ColorFormatter(colorGen, valueFormatter);
		}
		return formatter;
	}

	/**
		Get a formatter for nice formatting of annotation values, including units.
		A base formatter is created by {@link #getTypeFormatter(SessionControl, Type)}
		using the value type of the annotation. This formatter is wrapped by
		a {@link PrefixSuffixFormatter} if a unit is supplied or the annotation has
		a unit.
		
		@param sc The logged in user's session control
		@param a The annotation to get a formatter for
		@param unit The unit, or null to use the unit from the annotation 
			({@link Annotation#getUnit()}.
		@return A formatter object
		@since 2.9
	*/
	public static Formatter<?> getAnnotationFormatter(SessionControl sc, Annotation a, Unit unit)
	{
		Formatter<?> f = getTypeFormatter(sc, a.getValueType());
		return a.getUnitFormatter(f, unit);
	}
	
	/**
		Get a formatter that knows how to format results from formulas. This method 
		uses the value type of the formula ({@link Formula#getValueType()}) and 
		{@link #getTypeFormatter(SessionControl, Type)} to create the basic formatter.
		If the formula is numerical and uses coloring information the formatter is
		wrapped in a color formatter {@link #getColorFormatter(SessionControl, Coloring, Formatter)}.
		If the value type of the formula can't be determined a {@link ToStringFormatter}
		is used.
		
		@param sc The logged in user's session control
		@param formula The formula
		@return A formatter object
		@since 2.16
	*/
	@SuppressWarnings("unchecked")
	public static Formatter<?> getFormulaFormatter(SessionControl sc, Formula formula)
	{
		Formatter<?> f = null;
		Type valueType = formula.getValueType();
		if (valueType != null)
		{
			f = getTypeFormatter(sc, valueType);
			if (valueType.isNumerical()) 
			{
				f = getColorFormatter(sc, formula.getColoring(), (Formatter<Number>)f);
			}
		}
		else
		{
			f = new ToStringFormatter();
		}
		return f;
	}
	
}
