/*
	$Id: Tab.java 6576 2014-10-22 16:41:55Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.taglib.tab;

import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Map;

import javax.servlet.jsp.JspException;
import javax.servlet.jsp.JspTagException;
import javax.servlet.jsp.tagext.BodyTagSupport;
import javax.servlet.jsp.tagext.DynamicAttributes;

import net.sf.basedb.clients.web.extensions.DynamicActionAttribute;
import net.sf.basedb.clients.web.extensions.DynamicActionAttributes;
import net.sf.basedb.clients.web.extensions.tabcontrol.TabAction;
import net.sf.basedb.clients.web.taglib.DeprecationLogger;

/**
	This tag defines the title and contents of a single tab inside a tab control.
	<p>
	<b>Syntax:</b>
	<pre>
   &lt;t:tab
      id=...
      clazz=...
      style=...
      title=...
      tooltip=...
      validate=...
      activate=...
      visible=true|false
      helpid=...
      tabindex=...
   &gt;
</pre>

	<p>
	<table border=1 cellspacing=0 cellpadding=2>
	<tr>
		<td><b>Attribute</b></td>
		<td><b>Default value</b></td>
		<td><b>Required</b></td>
		<td><b>Description</b></td>
	</tr>
	<tr>
		<td>id</td>
		<td>tabcontrol1</td>
		<td>yes</td>
		<td>The ID of the tab. The value of this attribute
		goes directly into the standard HTML <code>id</code> attribute.
		This means that if you need a reference to the tab object in your own
		JavaScript code, you may for example use the following code:
		<p>
		<code>
		tab = document.getElementById('&lt;id&gt;');
		</code>
		</td>
	</tr>
	<tr>
		<td>clazz</td>
		<td>button</td>
		<td>no</td>
		<td>
			The value if this attribute goes directly into the standard HTML 
			<code>class</code> attribute. This allows you to apply different styles
			to different tables.
		</td>
	</tr>
	<tr>
		<td>style</td>
		<td>-</td>
		<td>no</td>
		<td>
			The value if this attribute goes directly into the standard HTML 
			<code>style</code> attribute. This allows you to apply different styles
			to different notes.
		</td>
	</tr>
	<tr>
		<td>title</td>
		<td>-</td>
		<td>yes</td>
		<td>The title of the tab.</td>
	</tr>
	<tr>
		<td>tooltip</td>
		<td>-</td>
		<td>no</td>
		<td>An optional tooltip that is shown as a popup text while the user
		is pointing the mouse over the tab's title. Not supported by all browsers.</td>
	</tr>
	<tr>
		<td>validate</td>
		<td>-</td>
		<td>no</td>
		<td>
		Deprecated. Use javascript and call TabControl.addTabValidator() to attach
		a validation function instead.
		An optional JavaScript expression that will be evaluated when
		the users tries to switch from this tab to another. If the expression returns 
		FALSE the switch will not be made. You may, for example, use this attribute 
		to validate input data in a form.
		</td>
	</tr>
	<tr>
		<td>activate</td>
		<td>-</td>
		<td>no</td>
		<td>
			Deprecated. Use javascript and call TabControl.addTabActivator() to attach
			a validation function instead.
			An optional JavaScript expression that will be executed when the switches
			to this tab.
		</td>
	</tr>
	<tr>
		<td>visible</td>
		<td>true</td>
		<td>no</td>
		<td>If this tab should be visible or not.
		</td>
	</tr>
	<tr>
		<td>helpid</td>
		<td></td>
		<td>no</td>
		<td>Optional ID for the help text associated with this tab. See the 
			{@link net.sf.basedb.clients.web.taglib.Help} taglib
			for more information.
		</td>
	</tr>
	<tr>
		<td>tabindex</td>
		<td>0</td>
		<td>no</td>
		<td>
			Tab index value to control the focus order. Use -1 to disable tabbing.
		</td>
	</tr>
	</table>

	@author Nicklas
	@version 2.0
	@see TabControl
*/
public class Tab
	extends BodyTagSupport
	implements TabAction, DynamicAttributes, DynamicActionAttributes
{

	/**
		
	*/
	private static final long serialVersionUID = -2308568821797105775L;

	/**
		The parent TabControl object.
	*/
	private TabControl tc;
	private Map<String, DynamicActionAttribute> dynamicAttributes;
	
	/**
		Optional <code>class</code> attribute.
		@since 2.17
	*/
	private String clazz;
	
	/**
		Optional <code>style</code> attribute
	*/
	private String style = null;

	/**
		The title of the tab.
	*/
	private String title = "";
	
	/**
		A popup tooltip text.
	*/
	private String tooltip = null;
	
	/**
		A JavaScript method that validates the tab content.
	*/
	@Deprecated
	private String validate = null;

	/**
		A JavaScript method that is invoked when activating the tab.
	*/
	@Deprecated
	private String activate = null;

	/**
		If the tab should be visible or not.
	*/
	private boolean visible = true;
	
	/**
		The help ID for this tab.
	*/
	private String helpid;
	
	private int tabIndex;

	
	/*
		Taglib initialization methods
	*/
	/**
		@since 2.17
	*/
	public void setClazz(String clazz)
	{
		this.clazz = clazz;
	}
	/**
		@since 2.17
	*/
	@Override
	public String getClazz()
	{
		return clazz;
	}
	
	public void setStyle(String style)
	{
		this.style = style;
	}
	@Override
	public String getStyle()
	{
		return style;
	}
	
	public void setTitle(String title)
	{
		this.title = title;
	}
	@Override
	public String getTitle()
	{
		return title;
	}
	
	public void setTooltip(String tooltip)
	{
		this.tooltip = tooltip;
	}
	@Override
	public String getTooltip()
	{
		return tooltip;
	}
	
	@Deprecated
	public void setValidate(String validate)
	{
		this.validate = validate;
		DeprecationLogger.warn("t:tab", "validate", "3.5");
	}
	@Override
	@Deprecated
	public String getOnValidate()
	{
		return validate;
	}

	/**
		@since 2.17
	*/
	@Deprecated
	public void setActivate(String activate)
	{
		this.activate = activate;
		DeprecationLogger.warn("t:tab", "activate", "3.5");
	}
	/**
		@since 2.17
	*/
	@Override
	@Deprecated
	public String getOnActivate()
	{
		return activate;
	}

	public void setVisible(boolean visible)
	{
		this.visible = visible;
	}
	@Override
	public boolean isVisible()
	{
		return visible;
	}

	public void setHelpid(String helpid)
	{
		this.helpid = helpid;
	}
	public String getHelpid()
	{
		return helpid;
	}

	public void setTabindex(int tabIndex)
	{
		this.tabIndex = tabIndex;
	}
	
	public int getTabindex()
	{
		return tabIndex;
	}
	
	/*
		From the DynamicAttributes interface
		------------------------------------
	*/
	@Override
	public void setDynamicAttribute(String uri, String name, Object value) 
		throws JspException 
	{
		if (value == null) return;
		if (dynamicAttributes == null)
		{
			dynamicAttributes = new LinkedHashMap<String, DynamicActionAttribute>();
		}
		dynamicAttributes.put(name, new DynamicActionAttribute(name, value.toString()));
	}
	// -------------------------------------
	@Override
	public Iterator<DynamicActionAttribute> getDynamicActionAttributes() 
	{
		return dynamicAttributes == null ? null : dynamicAttributes.values().iterator();
	}

	@Override
	public int doStartTag()
		throws JspException
	{
		if (!isVisible()) return SKIP_BODY;
		if (!(getParent() instanceof TabControl)) throw new JspTagException("Tag <t:tab> must be inside a <t:tabcontrol> tag");
		tc = (TabControl)getParent();
		return EVAL_BODY_BUFFERED;
	}

	@Override
	public int doEndTag()
		throws JspException
	{
		if (isVisible()) tc.addTab(this);
		return EVAL_PAGE;
	}

	/*
		From the TabAction interface
		----------------------------
	*/
	@Override
	public String getHelpId()
	{
		return helpid;
	}

	@Override
	public int getTabIndex()
	{
		return tabIndex;
	}
	
	@Override
	public String getContent()
	{
		return bodyContent != null ? bodyContent.getString() : null;
	}
	// -----------------------------
}


