/*
	$Id: CellValue.java 6127 2012-09-14 07:28:35Z nicklas $

	Copyright (C) 2006 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.taglib.table;

import javax.servlet.jsp.JspException;
import javax.servlet.jsp.JspTagException;
import javax.servlet.jsp.tagext.TagSupport;

import net.sf.basedb.util.formatter.Formatter;
import net.sf.basedb.util.formatter.ToStringFormatter;

/**
	This tag adds a value to a cell in a row. It should be placed inside a
	{@link Cell &lt;tbl:cell&gt;} tag.

	<p>
	<b>Syntax:</b>
	<pre>
   &lt;tbl:cellvalue
      value=...
      list=...
      separator=...
      prefix=...
      suffix=...
   &gt;
</pre>
	<p>
	<table border=1 cellspacing=0 cellpadding=2>
	<tr>
		<td><b>Attribute</b></td>
		<td><b>Default value</b></td>
		<td><b>Required</b></td>
		<td><b>Description</b></td>
	</tr>
	<tr>
		<td>value</td>
		<td>-</td>
		<td>no</td>
		<td>
			A value to display in the cell. If a {@link Formatter} has been
			defined for this column the value is formatted otherwise it is just
			converted to a string with the toString() method.
		</td>
	</tr>
	<tr>
		<td>list</td>
		<td>-</td>
		<td>no</td>
		<td>
			A list of values to display in the cell. Each value is formatted 
			the same way as for a single value. If a single value is also specified it is
			printed first.
		</td>
	</tr>
	<tr>
		<td>separator</td>
		<td>; </td>
		<td>no</td>
		<td>
			A separator to use between each value in a list.
		</td>
	</tr>
	<tr>
		<td>prefix</td>
		<td></td>
		<td>no</td>
		<td>
			A prefix that is added before each value. Since 2.9.
		</td>
	</tr>
	<tr>
		<td>suffix</td>
		<td></td>
		<td>no</td>
		<td>
			A suffix that is added after each value. Since 2.9.
		</td>
	</tr>
	</table>

	@author Nicklas
	@version 2.0
*/
public class CellValue
	extends TagSupport
{

	/**
		
	*/
	private static final long serialVersionUID = 4068184541728474047L;

	private static final Formatter toStringFormatter = new ToStringFormatter();
	
	/**
		The value to display.
	*/
	private Object value = null;
	private Iterable<?> list = null;
	private String prefix = "";
	private String suffix = "";
	private String separator = "; ";

	/*
		Taglib initialization methods
	*/	
	public void setValue(Object value)
	{
		this.value = value;
	}
	
	public Object getValue()
	{
		return value;
	}
	
	public void setList(Iterable<?> list)
	{
		this.list = list;
	}
	
	public Iterable<?> getList()
	{
		return list;
	}
	
	public void setSeparator(String separator)
	{
		this.separator = separator;
	}
	
	public String getSeparator()
	{
		return separator;
	}

	/**
		@since 2.9
	*/
	public void setPrefix(String prefix)
	{
		this.prefix = prefix == null ? "" : prefix;
	}
	/**
		@since 2.9
	*/
	public String getPrefix()
	{
		return prefix;
	}
	/**
		@since 2.9
	*/
	public void setSuffix(String suffix)
	{
		this.suffix = suffix == null ? "" : suffix;
	}
	/**
		@since 2.9
	*/
	public String getSuffix()
	{
		return suffix;
	}

	/*
		Display methods
	*/
	@Override
	@SuppressWarnings("unchecked")
	public int doStartTag()
		throws JspException
	{
		Cell cell =  (Cell)getParent();
		Formatter formatter = cell.getFormatter();
		if (formatter == null) formatter = toStringFormatter;
		
		StringBuilder output = new StringBuilder();
		String theSeparator = "";
		if (value != null)
		{
			output.append(prefix).append(formatter.format(value)).append(suffix);
			theSeparator = separator;
		}
		if (list != null)
		{
			for (Object o : list)
			{
				if (o != null)
				{
					output.append(theSeparator).append(prefix).append(formatter.format(o)).append(suffix);
					theSeparator = separator;
				}
			}
		}
		try
		{
			pageContext.getOut().print(output);
		}
		catch (Exception ex)
		{
			throw new JspTagException(ex.getMessage());
		}
		return SKIP_BODY;
	}

}


