/*
	$Id: BioMaterial.java 6127 2012-09-14 07:28:35Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

package net.sf.basedb.core;

import java.util.HashSet;
import java.util.Set;

import net.sf.basedb.core.Transactional.Action;
import net.sf.basedb.core.data.BioMaterialData;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;

/**
	This is the base class for for the three types of biomaterials:
	{@link BioSource}, {@link Sample} and {@link Extract}.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2012-09-14 09:28:35 +0200 (fr, 14 sep 2012) $
*/
public abstract class BioMaterial<D extends BioMaterialData>
	extends AnnotatedItem<D>
	implements Subtypable
{

	/**
		Get a biomaterial item when you know the id. This method can return any
		of the subclasses of {@link BioMaterial}. Eg. {@link BioSource},
		{@link Sample} or {@link Extract}.
		
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>Sample</code> item
		@throws ItemNotFoundException If an item with the specified 
			id is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have read permission to the item
		@throws BaseException If there is another error
		@since 2.17
	*/
	public static BioMaterial getById(DbControl dc, int id)
	{
		BioMaterial bm = dc.loadItemByDataClass(BioMaterial.class, BioMaterialData.class, id);
		if (bm == null) throw new ItemNotFoundException("BioMaterial[id="+id+"]");
		return bm;
	}
	
	/**
		The maximum length of the external id that can be stored in the database.
		@see #setExternalId(String)
	*/
	public static final int MAX_EXTERNAL_ID_LENGTH = BioMaterialData.MAX_EXTERNAL_ID_LENGTH;

	BioMaterial(D bioMaterialData)
	{
		super(bioMaterialData);
	}
	
	/*
		From the Subtypable interface
		-----------------------------
	*/
	@Override
	public ItemSubtype getItemSubtype()
	{
		return getDbControl().getItem(ItemSubtype.class, getData().getItemSubtype());
	}
	@Override
	public void setItemSubtype(ItemSubtype subtype)
	{
		checkPermission(Permission.WRITE);
		if (subtype != null)
		{
			subtype.setOnItem(this);
		}
		else
		{
			getData().setItemSubtype(null);
		}
	}
	// -------------------------------------------
	/*
		From the Annotatable interface
		-------------------------------------------
	*/
	/** 
		Get all child biomaterial items	with a subtype that 
		has the {@link ItemSubtype#getPushAnnotations()} flag set.
		@since 3.1
	*/
	@Override
	public Set<Annotatable> getAnnotatableParents()
		throws BaseException
	{
		Set<Annotatable> annotatable = new HashSet<Annotatable>();
		
		SpecialQuery<BioMaterialEventSource> childQuery = getChildCreationEvents();
		childQuery.join(Hql.innerJoin(null, "event", "evt", true));
		
		// BioMaterial
		childQuery.join(Hql.leftJoin("evt", "bioMaterial", "bm", null, true));
		childQuery.join(Hql.leftJoin("bm", "itemSubtype", "sub1", null, false));
		
		// Physical bioassays
		childQuery.join(Hql.leftJoin("evt", "physicalBioAssay", "pba", null, true));
		childQuery.join(Hql.leftJoin("pba", "itemSubtype", "sub2", null, false));
		
		childQuery.restrict(Restrictions.or(
			Restrictions.eq(Hql.property("sub1", "pushAnnotations"), Expressions.bool(true)),
			Restrictions.eq(Hql.property("sub2", "pushAnnotations"), Expressions.bool(true))
		));
		
		for (BioMaterialEventSource src : childQuery.list(getDbControl()))
		{
			BioMaterialEvent event = src.getEvent();
			BioMaterialEvent.Type eventType = event.getEventType();
			try
			{
				if (eventType == BioMaterialEvent.Type.CREATION)
				{
					BioMaterial bm = event.getBioMaterial();
					if (bm != null)	annotatable.add(bm);
				}
				else if (eventType == BioMaterialEvent.Type.BIOASSAY)
				{
					PhysicalBioAssay pba = src.getEvent().getPhysicalBioAssay();
					if (pba != null) annotatable.add(pba);
				}
			}
			catch (PermissionDeniedException ex)
			{}
		}
		
		return annotatable;
	}
	// -------------------------------------------

	/*
		From the BasicItem class
		------------------------
	 */
	/**
		On delete action: issue an update to decrease the size of all 
		{@link BioMaterialList}:s this is a member of.
	*/
	@Override
	void onBeforeCommit(Action action)
		throws NotLoggedInException, BaseException
	{
		super.onBeforeCommit(action);
		if (action == Action.DELETE)
		{
			org.hibernate.Session session = getDbControl().getHibernateSession();
			org.hibernate.Query query = HibernateUtil.getPredefinedSQLQuery(session, 
				"UPDATE_BIOMATERIALLIST_SIZE");
			/*
				UPDATE [BioMaterialLists] 
				SET [size] = [size] + :delta
				WHERE [id] IN (
					SELECT bml.[list_id]
					FROM [BioMaterialListMembers] bml
					WHERE bml.[biomaterial_id] = :bioMaterialId
				)
			*/
			query.setInteger("delta", -1);
			query.setInteger("bioMaterialId", getId());
			HibernateUtil.executeUpdate(query);
		}
	}
	// ------------------------------------
	
	/**
		Get the external id of the biomaterial. This value can be used to link
		with information in external databases. It is not used by the BASE core
		and it doesn't have to be unique.
		@return The external id for the biomaterial
	*/
	public String getExternalId()
	{
		return getData().getExternalId();
	}
	/**
		Set the external id of the biomaterial. The value
		may be null but must not be longer than the value specified by
		the {@link #MAX_EXTERNAL_ID_LENGTH} constant.
		@param externalId The new value for the external id
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws InvalidDataException If the string is too long
	*/
	public void setExternalId(String externalId)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		getData().setExternalId(StringUtil.setNullableString(externalId, "externalId", MAX_EXTERNAL_ID_LENGTH));
	}

	
	/**
		Count the number of already existing children to this biomaterial.
		The count includes items that the logged in user doesn't have access to.
		@param pooled If TRUE, the count is for biomaterials of the same type,
			if FALSE the count is for biomaterials of the child type
		@since 2.16
	*/
	public abstract long countChildren(boolean pooled);

	/**
		Get a query returning all events where this biomaterial has been used
		as a source item.
		@return A special query
		@since 3.0
	*/
	public SpecialQuery<BioMaterialEventSource> getChildCreationEvents()
	{
		SpecialQuery<BioMaterialEventSource> query = 
			new SpecialQuery<BioMaterialEventSource>(null, "BioMaterialEventSourceData", "es", 
					new BioMaterialEventSource.BioMaterialEventSourceTransformer());
		query.restrictPermanent(Restrictions.eq(Hql.property("es", "bioMaterial"), Hql.entity(this)));
		return query;
	}
	
}



