/*
	$Id: ExtendedProperty.java 6465 2014-06-04 12:41:12Z nicklas $

	Copyright (C) 2005 Samuel Andersson, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.text.NumberFormat;
import java.util.List;
import java.util.regex.Pattern;


/**
	Holds information about an extended property for a
	class.

	@see net.sf.basedb.core.data.ExtendableData
	@author Nicklas, Samuel
	@version 2.0
*/
public class ExtendedProperty
{
	
	/**
		A regexp checking for invalid characters.
	*/
	private static final Pattern valid = Pattern.compile("[a-zA-Z_][a-zA-Z0-9_]*");

	/**
		Check that the name only contains a-zA-Z0-9_ and starts with
		a letter or underscore.
	 	@param name Name to check 
	 	@return TRUE if valid, FALSE otherwise
		@since 2.4
	*/
	public static boolean isValidName(String name)
	{
		return name == null ? false : valid.matcher(name).matches();
	}
	
	private final String name;
	private final Type type;
	private final String column;
	private final String title;
	private final String description;
	private final int length;
	private final boolean nullable;
	private final boolean insertable;
	private final boolean updateable;
	private final boolean restrictedEdit;
	private final Formula.AverageMethod averageMethod;
	private final List<ExtendedPropertyLinker> linkers;

	/**
		Create a new extended property.
		See the getter methods for a description of the parameters.
	*/
	ExtendedProperty(String name, String title, String description, String column,
		Type type, int length, boolean nullable, boolean insertable, boolean updateable, 
		boolean restrictedEdit,
		Formula.AverageMethod averageMethod, List<ExtendedPropertyLinker> linkers)
	{
		this.name = name;
		this.title = title;
		this.description = description;
		this.column = column;
		this.type = type;
		this.length = length;
		this.nullable = nullable;
		this.insertable = insertable;
		this.updateable = updateable;
		this.restrictedEdit = restrictedEdit;
		this.averageMethod = averageMethod;
		this.linkers = linkers;
	}

	/**
		Get the name of the extended property. This is the name that is
		used in queries and for the <code>name</code> parameter of the
		<code>getExtended/setExtended</code> methods to get or
		set the value of the property of an <code>Extendable</code> item.
		@see net.sf.basedb.core.data.ExtendableData#getExtended(String) ExtendableData.getExtended
		@see net.sf.basedb.core.data.ExtendableData#setExtended(String,Object) ExtendableData.setExtended
		@return A string with the name
	*/
	public String getName()
	{
		return name;
	}
	/**
		Get the title of the extended property. The title is not used
		by the core but can be used by client applications for example
		to display a column header.
		@return A string with the title
	*/
	public String getTitle()
	{
		return title;
	}
	/**
		Get a description of the extended property. The description
		is not used by the core but can be used by client applications.
		@return A string with the description
	*/
	public String getDescription()
	{
		return description;
	}
	/**
		Get the database column of the extended property. This
		is the name of the column in the database which holds the
		value.
	 	@return a {@link java.lang.String} 
	*/
	public String getColumn()
	{
		return column;
	}
	/**
		Get the type of the extended property. The value returned is one
		of the defined constants by the {@link Type} enumeration.
	 	@return a Type enumeration. 
	*/
	public Type getType()
	{
		return type;
	}
	/**
		Get the maximum allowed length of the property value. This is only
		useful if it a string property.
		@return The maximum allowed length
	*/
	public int getLength()
	{
		return length;
	}
	/**
		Check if the property allows null values or not.
		@return TRUE if null values are allowed, FALSE otherwise
	*/
	public boolean isNullable()
	{
		return nullable;
	}
	/**
		Check if the value of this property should be inserted into the datbase
		when creating new items.
		@return TRUE if the value should be inserted, FALSE otherwise
	*/
	public boolean isInsertable()
	{
		return insertable;
	}
	/**
		Check if the value of this property should be updated in the database
		when an item is updated.
		@return TRUE if the value should be updated, FALSE otherwise
	*/
	public boolean isUpdateable()
	{
		return updateable;
	}
	
	/**
		Check if the value of this property is restricted for editing or not.
		This typically means that the logged in user must have a specific permission
		in order to change the value. To begin with restricted editing is only
		used for {@link User} items to lock some properties for non-administrator
		editing.
		@since 3.3
	*/
	public boolean isRestrictedEdit()
	{
		return restrictedEdit;
	}

	/**
		If it makes sense to take the average of multiple values for this
		property or not. For example, it doesn't make sense to average over
		position coordinates, but it does make sense to average over intensity
		values. By default all numerical values are averagable.
		@return TRUE if it makes sense to take the average, FALSE otherwise
		@see #getAverageMethod()
	*/
	public boolean isAveragable()
	{
		return averageMethod != null && averageMethod != Formula.AverageMethod.NONE;
	}
	
	/**
		Which method to use when calculating the average of a set of values.
		Prior to 2.4 only arithmetic mean was supported.
		@return The average method to use
		@since 2.4
	*/
	public Formula.AverageMethod getAverageMethod()
	{
		return averageMethod;
	}
	
	/**
		Parse a string and return an object of the correct type for this property.
		@param value The value to parse
		@return An object
		@throws InvalidDataException If the string cannot be converted to the correct type
		@see #getType()
	*/
	public Object parseString(String value)
		throws InvalidDataException
	{
		return getType().parseString(value);
	}

	/**
		Parse a string and return an object of the correct type for this property.
		Numeric properties are parsed with the specified number format. If the
		nullIfException parameter is TRUE null is returned if the string can't be
		parsed. Otherwise an exception is thrown.
		@param value The value to parse
		@param numberFormat The number format, or null to use Double.valueOf
		@param nullIfException TRUE to return null in case the string can't be parsed,
			FALSE to throw an exception
		@return An object or null
		@throws InvalidDataException If the string cannot be converted to the correct type
			and the nullIfException parameter is false
		@see #getType()
		@since 2.4
	*/
	public Object parseString(String value, NumberFormat numberFormat, boolean nullIfException)
	{
		return getType().parseString(value, numberFormat, nullIfException);
	}
	
	/**
		Check if an object is valid according to the specifications of this property.
		The checks include checking for null values if the property not {@link #isNullable()},
		check that the object is of the correct {@link #getType()} and the maximum 
		{@link #getLength()} of a string.
		@param value The value to check
		@throws InvalidDataException
	*/
	public void validateValue(Object value)
		throws InvalidDataException
	{
		if (value == null)
		{
			if (!isNullable()) throw new InvalidUseOfNullException(getName());
		}
		else
		{
			if (!getType().isCorrectType(value))
			{
				throw new InvalidDataException("Parameter "+getName()+" is a "+
					value.getClass()+", not a "+getType());
			}
			if (getType() == Type.STRING)
			{
				if (((String)value).length() > getLength())
				{
					throw new StringTooLongException(getName(), (String)value, getLength());
				}
			}
		}
	}
	
	/**
		Create a link for this extended property. This method checks the
		linkers in the order they are defined. The first one to return a non-null
		value is the winner.
		@param value The value to link
		@return The URL, or null if no matching linker could be found 
	*/
	public String getUrl(Object value)
	{
		if (value == null) return null;
		String url = null;
		if (linkers != null)
		{
			for (ExtendedPropertyLinker linker : linkers)
			{
				url = linker.getUrl(value);
				if (url != null) return url;
			}
		}
		return url;
	}
}

