/*
	$Id$

	Copyright (C) 2005 Johan Enell, Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson, Gregory Vincic
	Copyright (C) 2007 Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.ArrayDesignData;
import net.sf.basedb.core.data.ArrayDesignBlockData;
import net.sf.basedb.core.data.FeatureData;
import net.sf.basedb.core.data.ReporterData;
import net.sf.basedb.core.data.WellData;
import net.sf.basedb.core.hibernate.ExecuteUpdateWork;

import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.util.HashSet;
import java.util.Set;
import java.util.TreeSet;

/**
	Batcher class for Features or AffyFeatures. A single instance can be used
	for features belonging to a single {@link ArrayDesign}.
	<p>
	Create new {@link FeatureData} objects with either
	{@link #newFeature(ArrayDesignBlock, ReporterData)} or
	{@link #newFeature(ArrayDesignBlock, Well)}. Use the latter method if the array
	desing is connected to plates.
	<p>
	Depending on the {@link FeatureIdentificationMethod} used by the array design,
	you must set different properties on the <code>FeatureData</code> object.
	
	<ul>
	<li>NONE: Not supported
	<li>COORDINATES: {@link FeatureData#setRow(int)} and {@link FeatureData#setColumn(int)}
	<li>POSITION: {@link FeatureData#setPosition(int)}
	<li>FEATURE_ID: {@link FeatureData#setExternalId(String)}
	</ul>
	
	For all methods, each feature that is inserted must have a unqiue value. For example,
	there can't be two features with the same external ID if the FEATURE_ID method is 
	used. Values for the methods that are not used can still be set and saved to the
	database, but the uniqueness is not checked. The position number will be auto-generated 
	(and overwritten) if the identification method is not POSITION.
	
	@base.modified $Date$
	@author Enell, Nicklas
	@version 2.0
*/
public class FeatureBatcher
	extends BasicBatcher<FeatureData>
{

	/**
		Create a new batcher for regular features.
		
		@param dc The DbControl to use for database access and 
			permission checking
		@param arrayDesign The array design this batcher should 
			add features to
		@throws BaseException If there is an error
		@see ArrayDesign#getFeatureBatcher(FeatureIdentificationMethod, Job)
	*/
	static FeatureBatcher getNew(DbControl dc, ArrayDesign arrayDesign)
		throws DatabaseException, BaseException
	{
		FeatureBatcher fb = new FeatureBatcher(dc, arrayDesign);
		fb.initPermissions(0, 0);
		return fb;
	}


	/**
		The array design this batcher adds features to.
	*/
	private final ArrayDesign arrayDesign;
	
	/**
		Just a copy of the data object for the array design.
	*/
	private final ArrayDesignData arrayDesignData;
	
	private final FeatureIdentificationMethod fiMethod;
	
	/**
		Holds identification of already inserted features.
	*/
	private Set<Object> featureIds;
	
	/**
		Keeps track of the position number.
	*/
	private int currentPosition = 0;

	@SuppressWarnings("unchecked")
	FeatureBatcher(DbControl dc, ArrayDesign arrayDesign) 
		throws BaseException
	{
		super(dc.getHibernateSession(), FeatureData.class, null);
		this.arrayDesign = arrayDesign;
		this.arrayDesignData = arrayDesign.getData();
		this.fiMethod = arrayDesign.getFeatureIdentificationMethod();
		this.featureIds = fiMethod.caseInsensitive() && HibernateUtil.getDbEngine().caseInsensitiveComparison() ?
			new TreeSet(String.CASE_INSENSITIVE_ORDER) : new HashSet();
		setDbControl(dc);
	}

	/*
		From the BasicBatcher class
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return Item.FEATURE;
	}
	
	/**
		Validate all properties for the feature object:
		<ul>
		<li>The data object is not null
		<li>An array design block has been specfied
		<li>The postion number is > 0
		<li>The external id is not null or too long
		<li>The feature identification is unique
		<li>The data object belongs to the correct array design
		<li>The reporter match the well's reporter if a well has been specified
		</ul>
	*/
	@Override
	public void validate(FeatureData data) 
		throws InvalidDataException
	{
		if (data == null) throw new InvalidUseOfNullException("data");

		// Array design block must be set
		ArrayDesignBlockData adb = (ArrayDesignBlockData)getPropertyValue(data, "arrayDesignBlock");
		if (adb == null) throw new InvalidUseOfNullException("arrayDesignBlock");
		
		// Position must be > 0
		if (data.getPosition() <= 0)
		{
			throw new NumberOutOfRangeException("position", data.getPosition(), 1, false);
		}

		// External ID must not be too long
		data.setExternalId(StringUtil.setNullableString(data.getExternalId(), 
			"externalId", FeatureData.MAX_EXTERNAL_ID_LENGTH));

		// Feature identifier must be unique
		Object featureId = fiMethod.getIdentifier(data);
		if (featureIds.contains(featureId))
		{
			throw new ItemAlreadyExistsException("Another feature with the same identifier already exists: " + 
				fiMethod.toString() + "=" + featureId);
		}
		featureIds.add(featureId);

		// verify that the data object belongs to the correct raw bioassay
		if (!arrayDesignData.equals(adb.getArrayDesign()))
		{
			throw new InvalidDataException("Feature data belongs to a different array design");
		}
		
		// If a well is specified, verify that the reporter is the same
		WellData wd = (WellData)getPropertyValue(data, "well");
		if (wd != null)
		{
			ReporterData rd = data.getReporter();
			ReporterData wr = wd.getReporter();
			if ((rd != null && !rd.equals(wr)) || (rd == null && wr != null))
			{
				throw new InvalidDataException("The reporter in the well and on the feature doesn't match.");
			}
		}
	}

	/**
		Set the position number and add the feature to the insert queue.
	*/
	@Override
	public void insert(FeatureData data)
		throws BaseException
	{
		if (fiMethod != FeatureIdentificationMethod.POSITION)
		{
			data.setPosition(++currentPosition);
		}
		super.insert(data);
	}

	/**
		Always throws a <code>PermissionDeniedException</code>, since a feature cannot
		be modifed once it has been created.
	*/
	@Override
	public void update(FeatureData data)
		throws BaseException
	{
		throw new PermissionDeniedException(Permission.WRITE, "Feature");
	}

	/**
		Always throws a <code>PermissionDeniedException</code>, since a feature cannot
		be modifed once it has been created.
		@see #deleteAll()
	*/
	@Override
	public void delete(int id)
		throws BaseException
	{
		throw new PermissionDeniedException(Permission.DELETE, "Feature");
	}

	/**
		Grant read permission if the logged in user has read permission
		to the array design. Grant write, delete and create permission
		if the logged in user has write permission to the array design.
	*/
	@Override
	void initPermissions(int granted, int denied)
		throws BaseException
	{
		if (arrayDesign.hasPermission(Permission.READ))
		{
			granted |= Permission.grant(Permission.READ);
		}
		if (arrayDesign.hasPermission(Permission.WRITE))
		{
			granted |= Permission.grant(Permission.WRITE, Permission.DELETE, Permission.CREATE);
		}
		super.initPermissions(granted, denied);
	}
	
	@Override
	PluginPermission getPluginPermissions()
	{
		// Return NO_PLUGIN, since all permissions are taken care of by the initPermissions method
		return PluginPermission.NO_PLUGIN;
	}
	
	/**
		Set the hasFeature flag on the array design if features has been added.
	*/
	@Override
	void onBeforeClose() 
		throws BaseException
	{
		int insertCount = getTotalInsertCount();
		if (insertCount > 0)
		{
			arrayDesignData.setHasFeatures(true);
			arrayDesignData.setNumDbFeatures(insertCount);
		}
	}
	// -------------------------------------------

	/**
		Get the array design this batcher is used for.
	 	@return An {@link net.sf.basedb.core.ArrayDesign} 
	*/
	public ArrayDesign getArrayDesign()
	{
		return arrayDesign;
	}
	
	/**
		Create a new feature. 
	 	@param adb  The array design of the feature.
	 	@param well The well of the feature. Null is not allowed. The plate of the well must
	 		have a position on the ArrayDesign of the current FeatureBatcher.
		@return A feature data object
		@throws BaseException If there is an error
	*/
	public FeatureData newFeature(ArrayDesignBlock adb, Well well)
		throws BaseException
	{
		int index = arrayDesign.getPlatePosition(well.getPlate());
		if (index == -1)
		{
			throw new BaseException("Cannot find well "+well+" from plate "+well.getPlate()+" in arraydesign "+arrayDesign+".");
		}
		return new FeatureData(adb.getData(), well.getData());
	}

	/**
		Create a new feature. 
	 	@param adb
	 	@param reporter 
		@return A feature data object
		@throws BaseException If there is an error
	*/
	public FeatureData newFeature(ArrayDesignBlock adb, ReporterData reporter)
		throws BaseException
	{
		return new FeatureData(adb.getData(), reporter);
	}


	/**
		Delete all features belonging to this array design.
	*/
	void deleteAll()
		throws BaseException
	{
		String sql = PredefinedQuery.getQueryString("DELETE_FEATURES_FOR_ARRAYDESIGN");
		/**
			DELETE FROM Features f, ArrayDesignBlocks adb
			WHERE adb.id = f.arraydesignblock_id AND adb.arraydesign_id = ?
		*/
		try
		{
			if (debugSqlEnabled) logSql.debug("Deleting features: " + sql);
			HibernateUtil.doReturningWork(getDbControl().getHibernateSession(), 
				new ExecuteUpdateWork(getSessionControl(), sql)
				{
					@Override
					public void setParameters(PreparedStatement ps)
						throws SQLException
					{
						ps.setInt(1, arrayDesign.getId());
					}
					
				}
			);
			arrayDesignData.setHasFeatures(false);
		}
		catch (SQLException ex)
		{
			logSql.error(sql, ex);
			LogUtil.logSQLExceptionChain(logSql, ex);
			throw new BaseException(ex);
		}
	}
}
