/*
	$Id: Removable.java 4889 2009-04-06 12:52:39Z nicklas $

	Copyright (C) 2005 Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

/**
	A <code>Removable</code> item is an item that can be flagged
	as removed. This doesn't generally remove the information about the
	item from the database, but can be used by client applications
	to hide items that the user is not interested in.
	<p>
	For example, all methods generating lists of items, will by default
	not include those that are flagged as removed.
	<p>
	At regular intervals, a cleanup application is checking the database for 
	items that are flagged to be removed and deletes them permanently.
	<p>
	
	<b>Reference implementation</b><br>
	<pre class="code">
public boolean isRemoved()
{
   return getData().isRemoved();
}
public void setRemoved(boolean removed)
   throws PermissionDeniedException
{
   checkPermission(removed ? Permission.DELETE : Permission.WRITE);
   getData().setRemoved(removed);
}
</pre>
	@author Nicklas
	@version 2.0
	@base.modified $Date: 2009-04-06 14:52:39 +0200 (må, 06 apr 2009) $
*/
public interface Removable
	extends Identifiable
{

	/**
		Check if the removed flag is set for this item.
		@return TRUE if the item is flagged as removed, FALSE otherwise
	*/
	public boolean isRemoved();

	/**
		Set the removed flag for this item.
		@param removed TRUE if the item should be flagged as removed,
			FALSE otherwise
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#DELETE} permission for setting the flag 
			to TRUE or {@link Permission#WRITE} permission for setting the 
			flag to FALSE
	*/
	public void setRemoved(boolean removed)
		throws PermissionDeniedException;

}
