/*
	$Id $

	Copyright (C) 2011 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.Set;

import net.sf.basedb.core.data.ReporterClonePropertyData;
import net.sf.basedb.core.data.ReporterCloneTemplateData;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;

/**
	A reporter clone template is a definition of which reporter properties/annotations
	that should be cloned to a per-experiment table in the dynamic database. The
	cloning operation is initiated with {@link Experiment#getReporterCloneBatcher(DbControl)}. 
	The reporter internal and external id
	are always cloned and should not be part of the template. The cloning operation
	also creates a locked copy of the template, so that it is possible to make modifications
	to the original template without affecting the cloned data.

	@author Nicklas
	@since 3.1
	@base.modified $Date $
*/
public class ReporterCloneTemplate
	extends CommonItem<ReporterCloneTemplateData>
	implements Registered
	
{
	/**
		The type of item represented by this class.
		@see Item#REPORTERCLONETEMPLATE
		@see #getType()
	*/
	public static final Item TYPE = Item.REPORTERCLONETEMPLATE;
	
	/**
		Create a new reporter clone template item.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@return The new <code>ReporterCloneTemplate</code> item
		@throws BaseException If there is an error
	*/
	public static ReporterCloneTemplate getNew(DbControl dc)
		throws BaseException
	{
		ReporterCloneTemplate t = dc.newItem(ReporterCloneTemplate.class);
		t.setName("New reporter clone template");
		t.getData().setEntryDate(new Date());
		for (ClonableProperty cp : ClonableProperty.getAll())
		{
			if (cp.isMandatory()) t.addClonableProperty(cp);
		}
		return t;
	}
	
	/**
		Get a reporter clone template item when you know the id.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>ReporterCloneTemplate</code> item
		@throws ItemNotFoundException If an item with the specified 
			id is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static ReporterCloneTemplate getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		ReporterCloneTemplate t = dc.loadItem(ReporterCloneTemplate.class, id);
		if (t == null) throw new ItemNotFoundException("ReporterCloneTemplate[id="+id+"]");
		return t;
	}

	/**
		Get a query that returns reporter clone templates. It will always
		only return unlocked templates.
		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<ReporterCloneTemplate> getQuery()
	{
		ItemQuery<ReporterCloneTemplate> query = new ItemQuery<ReporterCloneTemplate>(ReporterCloneTemplate.class);
		query.restrictPermanent(
			Restrictions.eq(
					Hql.property("locked"), 
					Expressions.parameter("locked", false, Type.BOOLEAN)
			));
		return query;
	}

	ReporterCloneTemplate(ReporterCloneTemplateData data)
	{
		super(data);
	}

	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------

	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		Locked templates are READ-able by all, and everything else
		is denied. Unlocked templates use the regular permissions.
	*/
	@Override
	void initPermissions(int granted, int denied) 
		throws BaseException
	{
		if (isLocked())
		{
			granted |= Permission.grant(Permission.READ);
			denied |= Permission.deny(Permission.WRITE);
		}
		super.initPermissions(granted, denied);
	}
	
	/**
		
	*/
	@Override
	PluginPermission getPluginPermissions()
	{
		if (isLocked())
		{
			return new PluginPermission(Permission.grant(Permission.READ), Permission.deny(Permission.WRITE));
		}
		return super.getPluginPermissions();
	}
	
	/**
		Check if:
		<ul>
		<li>An experiment/virtualdb is using this template
		</ul>
	*/
	@Override
	public boolean isUsed()
		throws BaseException
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_VIRTUALDBS_FOR_REPORTERCLONETEMPLATE", "count(*)");
		/*
			SELECT {1}
			FROM VirtualDbData vdb
			WHERE vdb.reporterCloneTemplate = :reporterCloneTemplate
		*/
		query.setEntity("reporterCloneTemplate", this.getData());
		long count = HibernateUtil.loadData(Long.class, query);
		return count > 0 || super.isUsed();
	}

	/**
		Get all:
		<ul>
		<li>Experiment that is using this template
		<ul>
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();
		
		// Array batches
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_VIRTUALDBS_FOR_REPORTERCLONETEMPLATE", "vdb.experiment.id");
			/*
			SELECT {1}
			FROM VirtualDbData vdb
			WHERE vdb.reporterCloneTemplate = :reporterCloneTemplate
			*/
		query.setEntity("reporterCloneTemplate", this.getData());
		addUsingItems(using, Item.EXPERIMENT, query);		
		return using;
	}
	// -------------------------------------------

	/*
		From the Registered interface
		-----------------------------
	*/
	@Override
	public Date getEntryDate()
	{
		return DateUtil.copy(getData().getEntryDate());
	}
	// -------------------------------------------

	/**
		Allow the description to be changed with RESTRICTED_WRITE 
		permission.
	*/
	@Override
	public void setDescription(String description)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.RESTRICTED_WRITE);
		NameableUtil.setDescription(getData(), description);
	}
	
	/**
		If the template is locked it means that it is used by an experiment
		and can't be modified or deleted. Locked templates are managed 
		automatically by the BASE core.
	*/
	public boolean isLocked()
	{
		return getData().isLocked();
	}
	
	/**
		Get the number of bytes the cloned reporter information use in
		the database.
	*/
	public long getBytes()
	{
		return getData().getBytes();
	}
	void addBytes(long bytes)
	{
		getData().setBytes(bytes + getData().getBytes());
	}

	/**
		Add a clonable property to this template.
		@param property Information about the property to include in the 
			template (null is not allowed)
	*/
	public void addClonableProperty(ClonableProperty property)
	{
		checkPermission(Permission.WRITE);
		if (property == null) throw new InvalidUseOfNullException("property");
		getData().getProperties().put(property.getName(), property.toData());
	}
	
	/**
		Remove the property with the given name from this template
		@param propertyName The name of a property (null is not allowed)
	*/
	public void removeClonableProperty(String propertyName)
	{
		checkPermission(Permission.WRITE);
		if (propertyName == null) throw new InvalidUseOfNullException("propertyName");
		ClonableProperty cp = getClonableProperty(propertyName);
		if (cp != null)
		{
			if (cp.isMandatory())
			{
				throw new PermissionDeniedException("Property is mandatory: " + propertyName);
			}
			getData().getProperties().remove(propertyName);
		}
	}
	
	/**
		Get information about a clonable property that is included in this template.
		@param propertyName The name of a property
		@return Cloned property information, or null if the named property is not
			part of this template
	*/
	public ClonableProperty getClonableProperty(String propertyName)
	{
		ReporterClonePropertyData data = getData().getProperties().get(propertyName);
		return data == null ? null : new ClonableProperty(propertyName, data);
	}
	
	/**
		Check if this template contains a clonable property with
		the given name.
		@param propertyName The property name to check
		@return TRUE if the template contains the given property, FALSE if not
	*/
	public boolean hasClonableProperty(String propertyName)
	{
		return getData().getProperties().containsKey(propertyName);
	}
	
	/**
		Get a list with all clonable properties in this template. The returned list 
		is a copy of the information in this template and modifications to the list
		do not affect this template.
		@return A list with cloned properties
	*/
	public List<ClonableProperty> getClonableProperties()
	{
		return getClonableProperties(getData());
	}
	
	/**
		Create a deep clone of this template. Modifications made to the template
		after this call do not affect the copy. The returned copy is locked and
		cannot be modifed. Properties that are part of the original template,
		but no longer exists are not copied.
	*/
	ReporterCloneTemplateData createLockedCopy()
	{
		ReporterCloneTemplateData copy = new ReporterCloneTemplateData();
		copy.setName(getName());
		copy.setDescription(getDescription());
		copy.setOwner(getData().getOwner());
		copy.setLocked(true);
		copy.setEntryDate(new Date());
		
		Map<String, ReporterClonePropertyData> propertiesCopy = copy.getProperties();
		List<ClonableProperty> all = ClonableProperty.getAll();
		for (ClonableProperty cp : all)
		{
			if (hasClonableProperty(cp.getName()))
			{
				propertiesCopy.put(cp.getName(), cp.toData());
			}
		}
		return copy;
	}
	
	static ReporterCloneTemplateData createAllCopy(String name)
	{
		ReporterCloneTemplateData copy = new ReporterCloneTemplateData();
		copy.setName(name);
		copy.setLocked(true);
		copy.setEntryDate(new Date());
		
		Map<String, ReporterClonePropertyData> propertiesCopy = copy.getProperties();
		List<ClonableProperty> all = ClonableProperty.getAll();
		for (ClonableProperty cp : all)
		{
			propertiesCopy.put(cp.getName(), cp.toData());
		}
		return copy;
		
	}
	
	static List<ClonableProperty> getClonableProperties(ReporterCloneTemplateData template)
	{
		Set<Map.Entry<String, ReporterClonePropertyData>> properties = template.getProperties().entrySet();
		List<ClonableProperty> list = new ArrayList<ClonableProperty>(properties.size());
		for (Map.Entry<String, ReporterClonePropertyData> data : properties)
		{
			ClonableProperty cp = new ClonableProperty(data.getKey(), data.getValue());
			int insertIndex = Collections.binarySearch(list, cp);
			if (insertIndex < 0) list.add(-1-insertIndex, cp);
		}
		return list;
	}

	
}
