/*
	$Id: ReporterPropertyExpression.java 6127 2012-09-14 07:28:35Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.query.Query;
import net.sf.basedb.core.query.QueryElement;
import net.sf.basedb.core.query.QueryType;
import net.sf.basedb.core.query.Expression;
import net.sf.basedb.util.EqualsHelper;

import java.util.Collection;
import java.util.Iterator;
import java.util.Map;
import java.util.HashMap;

/**
	An expression representing the property of a
	reporter.
	<p>
	Throws an <code>UnsupportedOperationException</code> if 
	{@link Query#getQueryType()} isn't {@link QueryType#SQL}.

	@author Nicklas
	@version 2.0
	@see net.sf.basedb.core.query.Dynamic#reporter(String)
	@base.modified $Date: 2012-09-14 09:28:35 +0200 (fr, 14 sep 2012) $
*/
public class ReporterPropertyExpression
	implements Expression
{
	private static volatile Map<String, String> propertyToColumn;

	private final String property;
	
	private final String column;

	private final Boolean useCloned;
	
	/**
		Create a new reporter property expression. The expression 
		automatically selects if cloned reporter information
		should be used or not.
		@param property The name of the property
	*/
	public ReporterPropertyExpression(String property)
		throws InvalidDataException
	{
		this(property, null);
	}

	/**
		Create a new reporter property expression. The expression may
		be forced to use either the cloned or master reporter information.
		@param property The name of the property
		@param useCloned TRUE to use cloned reporter information (if available),
			FALSE to not use cloned information, null to automatically decide
		@since 3.1
	*/
	public ReporterPropertyExpression(String property, Boolean useCloned)
	{
		this.property = property;
		this.column = getColumn(property);
		this.useCloned = useCloned;
		if (column == null) throw new InvalidDataException("Unknown property '"+property+"' of reporter.");
	}
	
	/*
		From the QueryElement interface
		-------------------------------------------
	*/
	@Override
	public String toQl(Query query, DbControl dc)
		throws BaseException
	{
		if (query.getQueryType() == QueryType.SQL)
		{
			String joinAlias = RealTable.REPORTERS.getAlias();
			
			if (query instanceof ClonableReporterQuery)
			{
				ClonableReporterQuery clonableQuery = (ClonableReporterQuery)query;
				if (useCloned(clonableQuery)) 
				{
					if (!dc.loadItem(clonableQuery.getVirtualDb()).hasClonedReporterProperty(property))
					{
						throw new InvalidDataException("Reporter property is not cloned: " + property);
					}
					joinAlias = VirtualTable.CLONED_REPORTERS.getAlias();
				}
			}
			return HibernateUtil.quote(joinAlias) + "." + HibernateUtil.quote(column); 
		}
		else
		{
			throw new UnsupportedOperationException("'reporterProperty' operation not supported by "+query.getQueryType()+" query.");
		}
	}
	/**
		@return Always null
	*/
	@Override
	public Collection<? extends QueryElement> getChildren()
	{
		return null;
	}
	// -------------------------------------------
	/*
		From the Object class
		-------------------------------------------
	*/
	@Override
	public String toString()
	{
		return RealTable.REPORTERS.getAlias() + "." + property; 
	}
	/**
		This expression is equal to another ReporterPropertyExpression if they 
		represent the same reporter property.
	*/
	@Override
	public boolean equals(Object other)
	{
		if (this == other) return true;
		if (other == null || this.getClass() != other.getClass()) return false;
		ReporterPropertyExpression o = (ReporterPropertyExpression)other;
		return EqualsHelper.equals(property, o.property) && useCloned == o.useCloned;
	}
	@Override
	public int hashCode()
	{
		return EqualsHelper.hashCode(property) + (useCloned ? 1 : 0);
	}
	// -------------------------------------------
	
	/**
		Check if this reporter property should use the cloned
		reporter for the given query or not. If the query is using a
		virtual database that doesn't have cloned reporters, FALSE is 
		returned. Othersise, if an explicit request has been made to
		use/not use cloned reporters, that value is returned. If
		no explicit request has been made, the outcome depends
		on the {@link DynamicQuery#useClonedReporters()}.
		
		@param query The query to check with
		@return TRUE if the cloned reporter table should be used, FALSE if not
		@since 3.1
	*/
	public boolean useCloned(ClonableReporterQuery query)
	{
		// Check if the virtual database has cloned reporters
		if (!query.hasClonedReporters()) return false;
		// Explicit request to use/not use cloned reporters
		if (useCloned != null) return useCloned.booleanValue();
		// Auto-select based on query 
		return query.useClonedReporters();
	}
	
	private static String getColumn(String property)
	{
		if (propertyToColumn == null)
		{
			synchronized (ReporterPropertyExpression.class)
			{
				// So we are really sure no other thread created this while we were waiting
				if (propertyToColumn == null)
				{
					propertyToColumn = new HashMap<String, String>();
					org.hibernate.mapping.PersistentClass pc = HibernateUtil.getClassMapping("net.sf.basedb.core.data.ReporterData");
					Iterator propIt = pc.getPropertyIterator();
					while (propIt.hasNext())
					{
						org.hibernate.mapping.Property prop = (org.hibernate.mapping.Property)propIt.next();
						Iterator colIt = prop.getColumnIterator();
						if (colIt.hasNext())
						{
							org.hibernate.mapping.Column col = (org.hibernate.mapping.Column)colIt.next();
							propertyToColumn.put(prop.getName(), col.getName());
						}
					}
					propertyToColumn.put("id", "id"); //Add id as a property because hibernate wont
				}
			}
		}
		return propertyToColumn.get(property);
	}
}
