/*
	$Id: Update.java 6631 2014-11-27 12:06:57Z nicklas $

	Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg, Martin Svensson
	Copyright (C) 2007 Johan Enell, Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.hibernate.mapping.Table;

import net.sf.basedb.core.authentication.LoginRequest;
import net.sf.basedb.core.data.AnnotatableData;
import net.sf.basedb.core.data.AnnotationTypeData;
import net.sf.basedb.core.data.BioMaterialEventData;
import net.sf.basedb.core.data.BioPlateData;
import net.sf.basedb.core.data.BioPlateTypeData;
import net.sf.basedb.core.data.ChangeHistoryDetailData;
import net.sf.basedb.core.data.ContextData;
import net.sf.basedb.core.data.DerivedBioAssayData;
import net.sf.basedb.core.data.ExperimentData;
import net.sf.basedb.core.data.FileServerData;
import net.sf.basedb.core.data.FileSetMemberData;
import net.sf.basedb.core.data.ItemSubtypeData;
import net.sf.basedb.core.data.JobData;
import net.sf.basedb.core.data.MeasuredBioMaterialData;
import net.sf.basedb.core.data.PropertyFilterData;
import net.sf.basedb.core.data.ReporterCloneTemplateData;
import net.sf.basedb.core.data.RoleKeyData;
import net.sf.basedb.core.data.SchemaVersionData;
import net.sf.basedb.core.data.ShareableData;
import net.sf.basedb.core.dbengine.DbEngine;
import net.sf.basedb.core.dbengine.TableInfo;
import net.sf.basedb.core.dbengine.TableInfo.ForeignKeyInfo;
import net.sf.basedb.core.hibernate.TableInfoWork;
import net.sf.basedb.core.log.ChangeType;
import net.sf.basedb.util.MD5;
import net.sf.basedb.util.Values;

/**
	This class contains static methods used to update items already in the database.
	
	<table>
	<tr>
		<th>Schema version</th>
		<th>Description</th>
	</tr>
	
	<tr>
		<td>90</td>
		<td>
			The schema as released in BASE 2.17. For information about schema changes before
			this check the BASE 2.17 code.
		</td>
	</tr>
	
	<tr>
		<td>100</td>
		<td>
			The schema as released in BASE 3.0. We don't list all changes here.
		</td>
	</tr>
	<tr>
		<td>101 and 102</td>
		<td>
			Fixes an issue with duplicate {@link FileSetMemberData} entries
			due to a bug in the upgrade program. Make sure that the unique
			constraint that should prevent duplicates exists.
		</td>
	</tr>
	<tr>
		<td>103</td>
		<td>
			Added {@link ReporterCloneTemplateData} and related classes. No
			special update is required.
		</td>
	</tr>
	<tr>
		<td>104</td>
		<td>
			Make {@link ExperimentData}  {@link AnnotatableData}. No
			special update is required.
		</td>
	</tr>
	<tr>
		<td>105</td>
		<td>
			Added {@link ItemSubtypeData#getPushAnnotations()}. All existing subtypes
			are set to 'false'.
		</td>
	</tr>
	<tr>
		<td>106</td>
		<td>
			Added {@link BioPlateData#getEntryDate()} and {@link BioPlateData#getEventDate()}.
			All existing bioplates get a null value.
		</td>
	</tr>
	<tr>
		<td>107</td>
		<td>
			Added {@link BioPlateTypeData#getStorageType()}
			All existing bioplate types get a null value.
		</td>
	</tr>
	<tr>
		<td>108</td>
		<td>
			Added {@link BioPlateData#getSection()}, {@link BioPlateData#getTray()}
			and {@link BioPlateData#getPosition()}
			All existing bioplates get null values.
		</td>
	</tr>
	<tr>
		<td>109 and 110</td>
		<td>
			Added {@link DerivedBioAssayData#getPhysicalBioAssays()} and {@link DerivedBioAssayData#getParents()}
			and removed <code>DerivedBioAssayData.getPhysicalBioAssay()</code> and 
			<code>DerivedBioAssayData.getParent()</code>. Existing derived bioassays are updated in a two-step
			procedure. The first step is to copy existing information to the new tables and the second step is
			to drop columns that are no longer needed.
		</td>
	</tr>
	<tr>
		<td>111</td>
		<td>
			Not an actual database schema change, but we need to remove annotation sets that have been
			created without any annotations by the annotation importer and BASE 1->2 migration.
		</td>
	</tr>
	<tr>
		<td>112</td>
		<td>
			Not an actual database schema change, but we need to fix the parent_id and parent_type
			columns in the BioMaterials table to really reflect what is in the BioMaterialEventSources2 
			table. For more info see BASE ticket: http://base.thep.lu.se/ticket/1745
		</td>
	</tr>
	<tr>
		<td>113</td>
		<td>
			Made {@link Hardware}, {@link Software}, {@link Tag} and {@link BioMaterialList}
			{@link Annotatable} items. No special database update is needed. Only increase 
			the schema version.
		</td>
	</tr>
	<tr>
		<td>114</td>
		<td>
			Added {@link ChangeHistoryDetailData#getOldValue()} and {@link ChangeHistoryDetailData#getNewValue()}.
			An update that changes the log format for annotation changes is needed to make log entries
			using the same format. 
		</td>
	</tr>
	<tr>
		<td>115</td>
		<td>
			Added {@link AnnotationTypeData#getDisableLogOfValues()}. The update should set
			the value for all existing annotation types to false.
		</td>
	</tr>
	
	<tr>
		<td>116</td>
		<td>
			Added {@link JobData#getExternalId()} and made {@link JobData} a {@link ShareableData} item. 
			No special database update is needed. Only increase the schema version.
		</td>
	</tr>
	
	<tr>
		<td>117</td>
		<td>
			Added {@link FileServerData#getRootPath()} and {@link FileServerData#getSshFingerprint()}.
			No special database update is needed. Only increase the schema version.
		</td>
	</tr>
	<tr>
		<td>118</td>
		<td>
			Added {@link SchemaVersionData#getAppId()}.
			The update will set the app id for the current entry to 'net.sf.basedb.core'
			and the update the schema version.
		</td>
	</tr>
	</table>

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2014-11-27 13:06:57 +0100 (to, 27 nov 2014) $
*/
public final class Update
{
	/**
		Log core events.
	*/
	private static final org.slf4j.Logger log = 
		org.slf4j.LoggerFactory.getLogger(Update.class);

	/**
		Update the database by modifying existing items to follow new requirements.
		
		@param progress An object implementing the {@link ProgressReporter}
			interface
		@param rootLogin The root user login
		@param rootPassword The root user password
		@throws BaseException
	 */
	public static synchronized void updateDatabase(ProgressReporter progress, String rootLogin, String rootPassword)
		throws BaseException
	{
		if (progress != null) progress.display(0, "Updating database...");
		org.hibernate.Session session = null;
		try
		{
			Config.setProperty("extensions.disabled", "true");
			Application.start(false, false, false);

			// Test root user account
			SessionControl sc = Application.newSessionControl(null, null, null);
			LoginRequest loginRequest = new LoginRequest(rootLogin, rootPassword);
			sc.login(loginRequest);
			if (sc.getLoggedInUserId() != SystemItems.getId(User.ROOT))
			{
				throw new PermissionDeniedException("User '" + rootLogin + "' is not the root account.");
			}
			
			session = HibernateUtil.newSession();
			int schemaVersion = getSchemaVersionInTransaction(session, SchemaVersionData.BASE_APP_ID);
			final int total_progress_steps = 1 + Install.NEW_SCHEMA_VERSION - schemaVersion;
			final float progress_step = 100F / total_progress_steps;
			float progress_current = progress_step;
			
			// The 101 and 102 updates are only needed when updating from a BASE 3.0.0 installation
			if (schemaVersion < 101)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 101...");
				schemaVersion = updateToSchemaVersion101(session);
				progress_current += progress_step;
			}
			
			if (schemaVersion < 102)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 102...");
				schemaVersion = updateToSchemaVersion102(session);
				progress_current += progress_step;
			}
			
			if (schemaVersion < 106)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 106...");
				schemaVersion = setSchemaVersionInTransaction(session, 106);
				progress_current += 4 * progress_step;
			}
			
			if (schemaVersion < 108)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 108...");
				schemaVersion = setSchemaVersionInTransaction(session, 108);
				progress_current += 2 * progress_step;
			}
			
			if (schemaVersion < 109)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 110...");
				schemaVersion = updateToSchemaVersion110(session, schemaVersion);
				progress_current += progress_step;
			}
			
			if (schemaVersion < 111)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 111...");
				schemaVersion = updateToSchemaVersion111(session, progress);
				progress_current += progress_step;
			}
			
			if (schemaVersion < 112)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 112...");
				schemaVersion = updateToSchemaVersion112(session, progress);
				progress_current += progress_step;
			}

			if (schemaVersion < 113)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 113...");
				// Schemaversion 113 only updates the version number
				schemaVersion = setSchemaVersionInTransaction(session, 113);
				progress_current += progress_step;
			}

			if (schemaVersion < 114)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 114...");
				schemaVersion = updateToSchemaVersion114(session, progress);
				progress_current += progress_step;
			}
			
			if (schemaVersion < 117)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 117...");
				// Schemaversion 115 - 117 only updates the version number
				schemaVersion = setSchemaVersionInTransaction(session, 117);
				progress_current += 3 * progress_step;
			}
			
			if (schemaVersion < 118)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 118...");
				// Schemaversion 118 only updates the version number
				schemaVersion = setSchemaVersionInTransaction(session, 118);
				progress_current += progress_step;
			}

			sc.logout();
			if (progress != null) progress.display(100, "Database updated successfully.");
		}
		catch (BaseException ex)
		{
			if (progress != null) progress.display(100, "Database update failed: " + ex.getMessage()+"\n");
			throw ex;
		}
		finally
		{
			if (session != null) HibernateUtil.close(session);
			Application.stop();
		}
	}
	
	/**
		Verify the root user login and password. This method is called first 
		in the update procedure, which means that tables, etc. are usually
		still from the old version.
	*/
	static void verifyRootLoginAndPassword(String login, String password)
	{
	
		org.hibernate.Session session = null;
		org.hibernate.Transaction tx = null;
		
		try
		{
			session = HibernateUtil.newSession();
			tx = HibernateUtil.newTransaction(session);
			int schemaVersion = 0;
			try
			{
				org.hibernate.Query query = HibernateUtil.createQuery(session, 
						"SELECT sv.schemaVersion FROM SchemaVersionData sv");
				schemaVersion = HibernateUtil.loadData(Integer.class, query);
			}
			catch (Throwable t)
			{}
			
			if (schemaVersion > 0)
			{
				org.hibernate.Query query = HibernateUtil.createQuery(session, 
					"SELECT usr.id, usr.login FROM UserData usr WHERE usr.systemId='" + User.ROOT + "'");
				Object[] user = (Object[])query.uniqueResult();
				if (user == null)
				{
					throw new LoginException("User ROOT not found");
				}
				if (!login.equals(user[1]))
				{
					throw new LoginException("User '" + login + "' is not the ROOT user");
				}
				
				int id = (Integer)user[0];
				
				if (schemaVersion > 90)
				{
					// BASE 3
					query = HibernateUtil.createQuery(session, 
						"SELECT cryptedPassword FROM PasswordData pwd WHERE pwd.id="+id);
					String cryptedPassword = HibernateUtil.loadData(String.class, query);
					if (!User.checkPassword(password, cryptedPassword))
					{
						throw new LoginException("Incorrect password for user '" + login + "'");
					}
				}
				else if (schemaVersion == 90)
				{
					// BASE 2.17
					query = HibernateUtil.createSqlQuery(session, 
						"SELECT [md5password] FROM [Passwords] WHERE [id]="+id);
					String md5Password = HibernateUtil.loadData(String.class, query);
					if (!MD5.getHashString(password).equals(md5Password))
					{
						throw new LoginException("Incorrect password for user '" + login + "'");
					}
				}
				// We let all other conditions pass, since they will 
				// usually generate an error later
			}
		}
		finally
		{
			if (tx != null) HibernateUtil.rollback(tx);
			if (session != null) HibernateUtil.close(session);
		}
	}
	
	/**
		Get the current schema version number as it is stored in the database.
		@return The schema version number, or 0 if no schem version is found
	*/
	static int getSchemaVersionInTransaction(org.hibernate.Session session, String appId)
		throws BaseException
	{
		org.hibernate.Transaction tx = null;
		int schemaVersion = 0;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			schemaVersion = getSchemaVersion(session, appId);
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("getSchemaVersion: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}
	
	static int getSchemaVersion(org.hibernate.Session session, String appId)
		throws BaseException
	{
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "GET_SCHEMA_VERSION_FOR_APP");
		query.setString("appId", appId);
		SchemaVersionData data = HibernateUtil.loadData(SchemaVersionData.class, query);
		int schemaVersion = data == null ? 0 : data.getSchemaVersion();
		return schemaVersion;
	}

	/**
		Update the database with a new schema version number. This method should
		be called from an open transaction.
		
		@param schemaVersion The new schema version number
	 */
	static void setSchemaVersion(org.hibernate.Session session, String appId, int schemaVersion, int build)
		throws BaseException
	{
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "SET_SCHEMA_VERSION_FOR_APP");
		query.setInteger("schemaVersion", schemaVersion);
		query.setInteger("build", build);
		query.setString("appId", appId);
		int modified = HibernateUtil.executeUpdate(query);
		if (modified == 0)
		{
			// Insert
			SchemaVersionData sv = new SchemaVersionData();
			sv.setBuild(build);
			sv.setSchemaVersion(schemaVersion);
			sv.setAppId(appId);
			HibernateUtil.saveData(session, sv);
		}
	}
	
	private static void setSchemaVersion(org.hibernate.Session session, int schemaVersion)
		throws BaseException
	{
		setSchemaVersion(session, SchemaVersionData.BASE_APP_ID, schemaVersion, Application.getBuild());
	}
	
	/**
		Update the schema version version number using a new transaction.
		@param schemaVersion The new schema version number
	 */
	private static int setSchemaVersionInTransaction(org.hibernate.Session session, int schemaVersion)
	{
		org.hibernate.Transaction tx = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			setSchemaVersion(session, schemaVersion);
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("setSchemaVersion: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}
	
	/**
	 	Adjust the existing items in the database to be compatible with the latest mappings.
	 	No update to the schema version should be done here. The code must also  
	 	consider the case that the update failes at a later stage and that the same
	 	update is executed again.
	 	
	  	@param progress An object implementing the {@link ProgressReporter}
	 		interface
	  	@param rootLogin The root user login
	  	@param rootPassword The root user password
	  	@throws BaseException
	*/
	public static synchronized void adjustExistingItems(ProgressReporter progress, int schemaVersion, String rootLogin, String rootPassword)
		throws BaseException
	{
		org.hibernate.Transaction tx = null;
		org.hibernate.Session session = null;
		try
		{
			Config.setProperty("extensions.disabled", "true");
			Application.start(false, false, false);

			// Test root user account
			SessionControl sc = Application.newSessionControl(null, null, null);
			LoginRequest loginRequest = new LoginRequest(rootLogin, rootPassword);
			sc.login(loginRequest);
			if (sc.getLoggedInUserId() != SystemItems.getId(User.ROOT))
			{
				throw new PermissionDeniedException("User '" + rootLogin + "' is not the root account.");
			}
			
			session = HibernateUtil.newSession();
			tx = HibernateUtil.newTransaction(session);
			
			if (schemaVersion < 115)
			{
				// Set destroyed=false on all bioplates
				org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "SET_DISABLE_LOG_FOR_ANNOTATION_TYPES");
				/*
					UPDATE AnnotationTypeData at
					SET at.disableLogOfValues = false
					WHERE at.disableLogOfValues IS NULL
				*/
				query.executeUpdate();
			}
			
			if (schemaVersion < 118)
			{
				// Set appId=net.sf.basedb.core on the only existing schema version entry
				org.hibernate.Query query = HibernateUtil.createQuery(session, 
					"UPDATE SchemaVersionData sv " + 
					"SET sv.appId = :appId WHERE sv.appId IS NULL");
				query.setString("appId", SchemaVersionData.BASE_APP_ID);
				query.executeUpdate();
			}
			
			//	Commit the changes
			HibernateUtil.commit(tx);
			log.info("adjustExistingItems: OK");					
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			if (progress != null) progress.display(100, "The adjustment of the existing items failed: " + ex.getMessage()+"\n");
			log.info("adjustExistingItems: FAILED");
			throw ex;
		}
		finally
		{
			if (session != null) HibernateUtil.close(session);
			Application.stop();
		}
	}
	

	/**
	 	Remove duplicate file set members.
	 	
		@return The new schema version (=101)
	*/
	private static int updateToSchemaVersion101(org.hibernate.Session session)
		throws BaseException
	{
		final int schemaVersion = 101;
		org.hibernate.Transaction tx = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);

			org.hibernate.Query query = HibernateUtil.createQuery(session, 
				"SELECT fsm FROM FileSetMemberData fsm");
			List<FileSetMemberData> members = HibernateUtil.loadList(FileSetMemberData.class, query, null);
			Set<String> allMembers = new HashSet<String>();
			for (FileSetMemberData member : members)
			{
				int fileSetId = member.getFileSet().getId();
				int fileTypeId = member.getDataFileType().getId();
				int fileId = member.getFile().getId();
				String unique = fileSetId + "." + fileTypeId + "." + fileId;
				if (!allMembers.add(unique))
				{
					// This combination has been seen before
					session.delete(member);
				}
			}
			
			// Update the schema version number
			setSchemaVersion(session, schemaVersion);
	
			// Commit the changes
			HibernateUtil.commit(tx);
			log.info("updateToSchemaVersion101: OK");
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("updateToSchemaVersion101: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}

	/**
	 	Add unique index on FileSetMembers(fileset_id,datafiletype_id,file_id)
	 	
		@return The new schema version (=102)
	*/
	private static int updateToSchemaVersion102(org.hibernate.Session session)
		throws BaseException
	{
		final int schemaVersion = 102;
		try
		{
	
			// Try to drop just to make sure that we don't create duplicated indexes
			dropIndex(session, true, "FileSetMembers", "fileset_id", "datafiletype_id", "file_id");
			// Create the index
			createIndex(session, true, "FileSetMembers_uniquefile", "FileSetMembers", "fileset_id", "datafiletype_id", "file_id");
			
			// Update the schema version number
			setSchemaVersionInTransaction(session, schemaVersion);
	
			log.info("updateToSchemaVersion102: OK");
		}
		catch (SQLException ex)
		{
			log.error("updateToSchemaVersion102: FAILED", ex);
			throw new BaseException(ex);
		}
		catch (BaseException ex)
		{
			log.error("updateToSchemaVersion102: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}
	
	/**
	 	Move derived bioassay parent items to different tables. Get rid of
	 	old columns in the DerivedBioAssays table.
	 	
		@return The new schema version (=110)
	*/
	private static int updateToSchemaVersion110(org.hibernate.Session session, int currentSchemaVersion)
		throws BaseException
	{
		final int schemaVersion = 110;
		try
		{
			if (currentSchemaVersion < 109)
			{
				// First step is to copy existing information to the new tables
				org.hibernate.Transaction tx = null;
				try
				{
					tx = HibernateUtil.newTransaction(session);
					
					// Copy parent physical and derived bioassay information to the
					// new tables: ParentPhysicalBioAssays and ParentDerivedBioAssays
					org.hibernate.Query query = HibernateUtil.getPredefinedSQLQuery(session, "COPY_PARENT_PHYSICAL_BIOASSAYS");
					/*
						INSERT INTO ParentPhysicalBioAssays (physicalbioassay_id, derivedbioassay_id)
						SELECT bioassay_id, id FROM DerivedBioAssays
					*/
					query.executeUpdate();
					query = HibernateUtil.getPredefinedSQLQuery(session, "COPY_PARENT_DERIVED_BIOASSAYS");
					/*
						INSERT INTO ParentDerivedBioAssays (parentbioassay_id, derivedbioassay_id)
						SELECT parent_id, id FROM DerivedBioAssays WHERE NOT parent_id IS NULL
					*/
					query.executeUpdate();
					
					// Set is_root flag on derived bioassays
					query = HibernateUtil.getPredefinedSQLQuery(session, "SET_ISROOT_ON_DERIVED_BIOASSAYS");
					/*
						UPDATE DerivedBioAssays SET is_root = (parent_id IS NULL)
					*/
					query.executeUpdate();
					
					// Remove table filters, etc. that use the old properties
					cleanContextFromProperty(session, Item.DERIVEDBIOASSAY, "physicalBioAssay", null, true);
					cleanContextFromProperty(session, Item.DERIVEDBIOASSAY, "parent", null, true);
					
					setSchemaVersion(session, 109);
	
					// Commit the changes
					HibernateUtil.commit(tx);
				}
				finally
				{
					if (tx != null && !tx.wasCommitted()) HibernateUtil.rollback(tx);
				}
			}
			
			// Second step is to drop the old columns that are no longer used in BASE 3.2
			dropColumn(session, "DerivedBioAssays", "parent_id", null);
			dropColumn(session, "DerivedBioAssays", "bioassay_id", null);
				
			// Update the schema version number
			setSchemaVersionInTransaction(session, schemaVersion);
	
			log.info("updateToSchemaVersion110: OK");
		}
		catch (SQLException ex)
		{
			log.error("updateToSchemaVersion110: FAILED", ex);
			throw new BaseException(ex);
		}
		catch (RuntimeException ex)
		{
			log.error("updateToSchemaVersion110: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}

	/**
	 	Remove annotation sets that have no annotation values and no inherited annotations.
		@return The new schema version (=111)
	*/
	private static int updateToSchemaVersion111(org.hibernate.Session session, final ProgressReporter progress)
		throws BaseException
	{
		final int schemaVersion = 111;
		org.hibernate.Transaction tx = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			
			AnnotationSet.deleteEmptyAnnotationSets(session,
					new ProgressReporter() 
					{
						@Override
						public void display(int percent, String message) 
						{
							progress.append("\n\t  "+message);
						}
						
						@Override
						public void append(String message) 
						{
							progress.append(message);
						}
					}
				);
			
			// Update the schema version number
			setSchemaVersion(session, schemaVersion);
	
			// Commit the changes
			HibernateUtil.commit(tx);
			log.info("updateToSchemaVersion111: OK");
		}
		catch (RuntimeException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("updateToSchemaVersion111: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}

	/**
	 	Ensure that parent_id and parent_type in BioMaterials table is correct
	 	when compared to BioMaterialEventSources2 table.
		@return The new schema version (=112)
	*/
	private static int updateToSchemaVersion112(org.hibernate.Session session, final ProgressReporter progress)
		throws BaseException
	{
		final int schemaVersion = 112;
		org.hibernate.Transaction tx = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			
			// Get all BioMaterial that has null in parent_id and/or parent_type
			String hql = "select bm.creationEvent from MeasuredBioMaterialData bm" +
					" where bm.parent is null or bm.parentType is null";
			
			org.hibernate.Query query = HibernateUtil.createQuery(session, hql);
			List<BioMaterialEventData> events = HibernateUtil.loadList(BioMaterialEventData.class, query, null);
			int i = 0;
			for (BioMaterialEventData event : events)
			{
				BioMaterialEvent.ensureProductParentCorrectness(event);
				i++;
			}
			
			// Update the schema version number
			setSchemaVersion(session, schemaVersion);
	
			// Commit the changes
			HibernateUtil.commit(tx);
			log.info("updateToSchemaVersion112: OK");
		}
		catch (RuntimeException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("updateToSchemaVersion112: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}

	
	/**
	 	Ensure that old log entries for annotation changes use the same text format
	 	that is used after this update.
		@return The new schema version (=114)
	*/
	private static int updateToSchemaVersion114(org.hibernate.Session session, final ProgressReporter progress)
		throws BaseException
	{
		final int schemaVersion = 114;
		org.hibernate.Transaction tx = null;
		org.hibernate.StatelessSession statelessSession = null;
		
		try
		{
			tx = HibernateUtil.newTransaction(session);
			statelessSession = HibernateUtil.newStatelessSession(session);
			// Get all ChangeHistoryDetail entries that are related to annotation changes
			// (eg those that start with "Annotation create|updated|deleted:"
			String hql = "select ch from ChangeHistoryDetailData ch" +
					" where ch.changeInfo is not null";
			String count = "select count(*) from ChangeHistoryDetailData ch" +
					" where ch.changeInfo is not null";
			
			String fixSql = "update [ChangeHistoryDetails] set [change_info] = :ci, [change_type] = :ct where [id] = :id";
			org.hibernate.Query fixQuery = HibernateUtil.createSqlQuery(statelessSession, fixSql);

			org.hibernate.Query countQuery = HibernateUtil.createQuery(statelessSession, count);
			long numItems = HibernateUtil.loadData(Long.class, countQuery);
			
			org.hibernate.Query query = HibernateUtil.createQuery(statelessSession, hql);
			Iterator<ChangeHistoryDetailData> changes = HibernateUtil.loadIterator(ChangeHistoryDetailData.class, query, null);
			
			int i = 0;
			int delta = (int)numItems / 50;
			while (changes.hasNext())
			{
				ChangeHistoryDetailData change = changes.next();
				String info = change.getChangeInfo();
				HibernateUtil.evictData(session, change);
				boolean fixIt = true;
				if (info.startsWith("Annotation created: "))
				{
					// Annotation created: <name> --> ADD + Annotation[<name>]
					fixQuery.setString("ci", "Annotation[" + info.substring("Annotation created: ".length()) + "]");
					fixQuery.setInteger("ct", ChangeType.ADD.getValue());
				}
				else if (info.startsWith("Annotation updated: "))
				{
					// Annotation updated: <name> --> UPDATE + Annotation[<name>]
					fixQuery.setString("ci", "Annotation[" + info.substring("Annotation updated: ".length()) + "]");
					fixQuery.setInteger("ct", ChangeType.UPDATE.getValue());
				}
				else if (info.startsWith("Annotation deleted: "))
				{
					// Annotation deleted: <name> --> REMOVE + Annotation[<name>]
					fixQuery.setString("ci", "Annotation[" + info.substring("Annotation deleted: ".length()) + "]");
					fixQuery.setInteger("ct", ChangeType.REMOVE.getValue());
				}
				else if (info.startsWith("File created: "))
				{
					// File created: <name> --> ADD + File[<name>]
					fixQuery.setString("ci", "File[" + info.substring("File created: ".length()) + "]");
					fixQuery.setInteger("ct", ChangeType.ADD.getValue());
				}
				else if (info.startsWith("File updated: "))
				{
					// File updated: <name> --> UPDATE + File[<name>]
					fixQuery.setString("ci", "Annotation[" + info.substring("File updated: ".length()) + "]");
					fixQuery.setInteger("ct", ChangeType.UPDATE.getValue());
				}
				else if (info.startsWith("File deleted: "))
				{
					// File deleted: <name> --> REMOVE + File[<name>]
					fixQuery.setString("ci", "File[" + info.substring("File deleted: ".length()) + "]");
					fixQuery.setInteger("ct", ChangeType.REMOVE.getValue());
				}
				else if (info.startsWith("Updated: "))
				{
					// Updated: <list-of-properties> --> UPDATE + property (one entry for each in the list)
					String[] properties = info.substring("Updated: ".length()).split(", ");
					if (properties.length > 1)
					{
						for (int p = 1; p < properties.length; p++)
						{
							// Create new entries for the extra properties
							ChangeHistoryDetailData clone = new ChangeHistoryDetailData();
							clone.setChangeHistory(change.getChangeHistory());
							clone.setChangeInfo(properties[p]);
							clone.setChangeType(ChangeType.UPDATE.getValue());
							clone.setItemId(change.getItemId());
							clone.setItemType(change.getItemType());
							HibernateUtil.saveData(session, clone);
						}
						session.flush();
						session.clear();
					}
					fixQuery.setString("ci", properties[0]);
					fixQuery.setInteger("ct", ChangeType.UPDATE.getValue());
				}
				else
				{
					fixIt = false;
				}
				
				if (fixIt)
				{
					fixQuery.setInteger("id", change.getId());
					fixQuery.executeUpdate();
				}
				
				i++;
				if (progress != null && i % delta == 0) 
				{
					progress.append(".");
				}
				
			}
			
			// Update the schema version number
			setSchemaVersion(session, schemaVersion);
	
			// Commit the changes
			HibernateUtil.commit(tx);
			log.info("updateToSchemaVersion114: OK");
		}
		catch (RuntimeException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("updateToSchemaVersion114: FAILED", ex);
			throw ex;
		}
		finally
		{
			HibernateUtil.close(statelessSession);
		}
		return schemaVersion;
	}
	
	/**
	 	Vefify and update the remaining quantity of all biomaterials.
		@return The number of biomaterial fixed
	*/
	public static int recalculateRemainingQuantity(ProgressReporter progress)
		throws BaseException
	{
		org.hibernate.Session session = null;
		org.hibernate.Transaction tx = null;
		int numUpdated = 0;
		try
		{
			session = HibernateUtil.newSession();
			tx = HibernateUtil.newTransaction(session);
			
			// For temporary storing the new remaining quantities
			Map<Integer, Double> remaining = new HashMap<Integer, Double>(); 
			
			// Load used quantity for all events
			org.hibernate.Query eventQuery = HibernateUtil.createQuery(session, 
				"SELECT evt.bioMaterial.id, evt.usedQuantity FROM BioMaterialEventData evt WHERE NOT evt.usedQuantity IS NULL");
			
			// Load used quantity for sources to all events
			org.hibernate.Query sourcesQuery = HibernateUtil.createQuery(session, 
				"SELECT src.bioMaterial.id, src.usedQuantity FROM BioMaterialEventSourceData src WHERE NOT src.usedQuantity IS NULL");
	
			progress.display(10, "Loading biomaterial items...\n");
			
			List<Object[]> events = HibernateUtil.loadList(Object[].class, eventQuery, null);
			events.addAll(HibernateUtil.loadList(Object[].class, sourcesQuery, null));
	
			// Calculate the remaining quantity for each biomaterial
			// o[0] = biomaterial id, o[1] = used quantity			
			for (Object[] o : events)
			{
				Integer id = (Integer)o[0];
				Float usedQuantity = (Float)o[1];
				Double currentRemaining = remaining.get(id);
				if (currentRemaining == null) currentRemaining = 0.0;
				currentRemaining = currentRemaining - usedQuantity.doubleValue();
				remaining.put(id, currentRemaining);
			}
			progress.display(30, "Found "+ remaining.size() + " items. Checking remaining quantity...\n");
			
			// Update the remaining quantity
			org.hibernate.Query updateQuery = HibernateUtil.createQuery(session, 
				"UPDATE MeasuredBioMaterialData mbm " +
				"SET mbm.remainingQuantity = :remain "+
				"WHERE mbm = :bioMaterial");
			int numProcessed = 0;
			for (Map.Entry<Integer, Double> entry : remaining.entrySet())
			{
				MeasuredBioMaterialData bioMaterial = HibernateUtil.loadData(session, MeasuredBioMaterialData.class, entry.getKey());
				Double remainingQuantity = entry.getValue();
				Float currentRemaining = bioMaterial.getRemainingQuantity();
				if (currentRemaining == null || Math.abs(currentRemaining-remainingQuantity) > Math.abs(currentRemaining / 10000))
				{
					updateQuery.setEntity("bioMaterial", bioMaterial);
					updateQuery.setFloat("remain", remainingQuantity.floatValue());
					updateQuery.executeUpdate();
					numUpdated++;
				}
				numProcessed++;
				if (numProcessed % 100 == 0)
				{
					progress.display(30 + numProcessed * 60 / remaining.size(), "Processed " + numProcessed + " biomaterial items...                   ");
				}
			}
			progress.append("\n");
			if (numUpdated == 0)
			{
				progress.display(100, "Found " + numUpdated + " items to fix\n");
			}
			else
			{
				progress.display(95, "Found " + numUpdated + " items to fix. Comitting transaction...\n");
				// Commit the changes
				HibernateUtil.commit(tx);
				progress.display(100, "Completeted update of " + numUpdated + " items\n");
			}
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			throw ex;
		}
		return numUpdated;
	}

	

	private static TableInfo getTableInfo(org.hibernate.Session session, final String tableName)
		throws SQLException
	{
		return HibernateUtil.doReturningWork(session, new TableInfoWork(new Table(tableName)));
	}
	
	private static void dropColumn(org.hibernate.Session session, String tableName, String columnName, ProgressReporter progress)
		throws SQLException
	{
		org.hibernate.Transaction tx = null;
		org.hibernate.Query query = null;
		try
		{
			DbEngine engine = HibernateUtil.getDbEngine();
			
			log.debug("Dropping column: " + tableName + "." + columnName);
			tx = HibernateUtil.newTransaction(session);
			
			// Drop foreign keys where this column is used
			TableInfo info = getTableInfo(session, tableName);
			for (ForeignKeyInfo fk : info.getForeignKeys())
			{
				if (fk.getFkColumns().contains(columnName))
				{
					log.debug("Dropping foreign key: " + tableName + "." + fk.getName() + "=" + fk.getFkColumns());
					String sql = engine.getDropForeignKeySql(null, null, tableName, fk.getName());
					query = HibernateUtil.createSqlQuery(session, sql);
					query.executeUpdate();
				}
			}
			
			if (progress != null)
			{
				progress.display(90, "  --alter table " + tableName + " drop column " + columnName);
			}
			query = HibernateUtil.getPredefinedSQLQuery(session, "DROP_COLUMN", tableName, columnName);
			query.executeUpdate();
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}
	}

	private static void dropIndex(org.hibernate.Session session, boolean unique, String tableName, String... columnNames)
			throws SQLException
	{
		org.hibernate.Transaction tx = null;
		org.hibernate.Query query = null;
		try
		{
			log.debug("Dropping index: " + tableName + ".[" + Arrays.toString(columnNames) + "]");
			tx = HibernateUtil.newTransaction(session);
			
			TableInfo info = getTableInfo(session, tableName);
			String indexName = info.findIndexName(null, new HashSet<String>(Arrays.asList(columnNames)));
			if (indexName != null)
			{
				DbEngine engine = HibernateUtil.getDbEngine();
				String sql = engine.getDropIndexSql(null, null, tableName, indexName, unique);
				query = HibernateUtil.createSqlQuery(session, sql);
				query.executeUpdate();
			}
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}
	}

	private static void createIndex(org.hibernate.Session session, boolean unique, String indexName, String tableName, String... columnNames)
			throws SQLException
	{
		org.hibernate.Transaction tx = null;
		org.hibernate.Query query = null;
		try
		{
			log.debug("Creating index: " + tableName + ".[" + Arrays.toString(columnNames) + "]");
			tx = HibernateUtil.newTransaction(session);
			DbEngine engine = HibernateUtil.getDbEngine();
			String sql = engine.getCreateIndexSql(null, null, tableName, indexName, new LinkedHashSet<String>(Arrays.asList(columnNames)), unique);
			query = HibernateUtil.createSqlQuery(session, sql);
			query.executeUpdate();
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}
	}

	
	private static void dropTable(org.hibernate.Session session, String tableName, ProgressReporter progress)
	{
		org.hibernate.Transaction tx = null;
		org.hibernate.Query query = null;
		try
		{
			log.debug("Dropping table: " + tableName);
			progress.display(90, "  --drop table " + tableName);
			tx = HibernateUtil.newTransaction(session);
			query = HibernateUtil.getPredefinedSQLQuery(session, "DROP_TABLE", tableName);
			query.executeUpdate();
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}
	}
	
	
	/**
		Remove context-related information for the given item type.
	*/
	private static void removeContext(org.hibernate.Session session, int itemType)
	{
		org.hibernate.Transaction tx = null;
		org.hibernate.Query query = null;
		try
		{
			log.debug("Removing context for item: " + itemType);
			tx = HibernateUtil.newTransaction(session);
			
			query = HibernateUtil.createQuery(session, "SELECT ctx FROM ContextData ctx WHERE ctx.itemType = :itemType");
			query.setInteger("itemType", itemType);
			for (ContextData ctx : HibernateUtil.loadList(ContextData.class, query, null))
			{
				session.delete(ctx);
			}

			// DELETE FROM PluginDefinitionGuiContexts WHERE item_type=?
			query = HibernateUtil.getPredefinedSQLQuery(session, 
				"DELETE_PLUGINGUICONTEXTS_FOR_ITEMTYPE");
			query.setInteger("itemType", itemType);
			query.executeUpdate();
			
			log.debug("Removing role key for item: " + itemType);
			// DELETE FROM PluginKeys WHERE item_type=?
			query = HibernateUtil.getPredefinedSQLQuery(session, 
				"DELETE_PLUGINKEYS_FOR_ITEMTYPE");
			query.setInteger("itemType", itemType);
			query.executeUpdate();

			query = HibernateUtil.createQuery(session, "SELECT key FROM RoleKeyData key WHERE key.itemType = :itemType");
			query.setInteger("itemType", itemType);
			for (RoleKeyData key : HibernateUtil.loadList(RoleKeyData.class, query, null))
			{
				session.delete(key);
			}
			
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}
	}
	
	private static void cleanContextFromProperty(org.hibernate.Session session, Item itemType, 
		String propertyName, String replacementPropertyName, boolean forceRemoveFilter)
	{
		org.hibernate.Transaction tx = null;
		org.hibernate.Query query = null;
		try
		{
			log.debug("Cleaning context for item: " + itemType + "; property=" + propertyName + "; replacement=" + replacementPropertyName);
			tx = session.getTransaction().isActive() ? null : HibernateUtil.newTransaction(session);
			
			query = HibernateUtil.createQuery(session, "select ctx from ContextData ctx where ctx.itemType = " + itemType.getValue());
			List<ContextData> contexts = HibernateUtil.loadList(ContextData.class, query, null);
			
			for (ContextData ctx : contexts)
			{
				// Sort property
				if (ctx.getSortProperty() != null)
				{
					ctx.setSortProperty(replaceInCommaSeparatedList(ctx.getSortProperty(), propertyName, replacementPropertyName));
				}
				
				// "columns" settings
				Map<String, String> settings = ctx.getSettings();
				String columns = settings.get("columns");
				if (columns != null)
				{
					columns = replaceInCommaSeparatedList(columns, propertyName, replacementPropertyName);
					if (columns == null)
					{
						settings.remove("columns");
					}
					else
					{
						settings.put("columns", columns);
					}
				}
					
				// Filter
				Map<String, PropertyFilterData> filters = ctx.getPropertyFilters();
				List<String> filterProperties = new ArrayList<String>(filters.keySet());
				for (String property : filterProperties)
				{
					if (property.startsWith(propertyName))
					{
						PropertyFilterData old = filters.remove(property);
						if (replacementPropertyName != null)
						{
							filters.put(replacementPropertyName + property.substring(propertyName.length()), old);
						}
					}
				}
			}
			
			// Only commit if we started a new transaction
			if (tx != null) HibernateUtil.commit(tx);
		}
		catch (RuntimeException ex)
		{
			// Only rollback if we started the transaction
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}
		
	}
	
	
	private static String replaceInCommaSeparatedList(String list, String find, String replacement)
	{
		List<String> elements = new ArrayList<String>(Arrays.asList(list.split(",")));
		String result = list;
		
		Iterator<String> it = elements.iterator();
		boolean hasMatched = false;
		int index = 0;
		while (it.hasNext())
		{
			String value = it.next();
			if (value != null && value.startsWith(find))
			{
				// A match, replace or remove
				hasMatched = true;
				if (replacement != null)
				{
					elements.set(index, replacement + value.substring(find.length()));
				}
				else
				{
					it.remove();
				}
			}
		}
		if (hasMatched)
		{
			if (elements.size() == 0)
			{
				result = null;
			}
			else
			{
				result = Values.getString(elements, ",", true);
			}
		}
		return result;
	}

	

}
