/*
	$Id: Version.java 5607 2011-04-15 07:52:28Z nicklas $

	Copyright (C) 2006 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.Properties;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.io.IOException;
import java.net.URL;

import net.sf.basedb.util.FileUtil;
import net.sf.basedb.util.Values;

/**
	Get information about the BASE version.
	<p>
	The version information is stored in the <code>base.version</code> file. 

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2011-04-15 09:52:28 +0200 (fr, 15 apr 2011) $
*/
public class Version
{
	private static int major;
	private static int minor;
	private static int maintenance;
	private static int build;
	private static String suffix;
	
	static
	{
		init();
	}
	
	/**
		Loads the settings from the configuration file.
		@throws ConfigurationException This exception is thrown if the
			configuration settings cannot be loaded
	*/
	static synchronized void init()
		throws ConfigurationException
	{
		InputStream is = null;
		try
		{
			URL baseConfig = Version.class.getResource("/base.version");
			if (baseConfig == null)
			{
				throw new FileNotFoundException("/base.version");
			}
			is = baseConfig.openStream();
			Properties config = new Properties();
			config.load(is);
			major = Values.getInt(config.getProperty("major"));
			minor = Values.getInt(config.getProperty("minor"));
			maintenance = Values.getInt(config.getProperty("maintenance"));
			build = parseBuildNumber(config.getProperty("build"));
			suffix = config.getProperty("suffix", "");
		}
		catch (IOException ex)
		{
			Application.getLogger().warn("Error while loading version information from '/base.version'.", ex);
		}
		finally
		{
			if (is != null) FileUtil.close(is);
		}
	}
	
	/**
		Get the major version.
		@since 3.0 (was returning a String in 2.x)
	*/
	public static int getMajor()
	{
		return major;
	}
	
	/**
		Get the minor version.
		@since 3.0 (was returning a String in 2.x)
	*/
	public static int getMinor()
	{
		return minor;
	}
	/**
		Get the maintentance version.
		@since 3.0 (was returning a String in 2.x)
	*/
	public static int getMaintenance()
	{
		return maintenance;
	}
	
	/**
		Get the build version.
	*/
	public static int getBuild()
	{
		return build;
	}
	
	/**
		Get the version suffix string (eg "-dev" for development versions)
		@since 3.0
	*/
	public static String getSuffix()
	{
		return suffix;
	}
	
	public static String getVersion()
	{
		return major+"."+minor+"."+maintenance + suffix + "(build " + build + ")";
	}

	/**
		Compare the given version string, which should be of the
		form a.b or a.b.c, with the current BASE version. 
		If they are exactly equal 0 is returned.
		If the given string is less than the current version
		-1 is returned, if the given string is greater than
		the current version +1 is returned.
		
		@param version
		@return 0, -1 or +1
		@since 3.0
		@throws InvalidDataException If the given string isn't correctly formatted
	*/
	public static int compareWith(String version)
	{
		if (version == null) throw new NullPointerException("version");
		Pattern p = Pattern.compile("(\\d+)\\.(\\d+)\\.?(\\d*)");
		Matcher m = p.matcher(version);
		if (!m.matches()) 
		{
			throw new InvalidDataException("Version is not correctly formatted: " + version);
		}
		
		// Major and minor as a first step
		int vMajor = Values.getInt(m.group(1));
		int vMinor = Values.getInt(m.group(2));
		if (vMajor != major) return vMajor < major ? -1 : 1;
		if (vMinor != minor) return vMinor < minor ? -1 : 1;
		
		// Need to check maintenance as well (if given)
		if (m.groupCount() > 2)
		{
			int vMaintenance = Values.getInt(m.group(3));
			if (vMaintenance != maintenance) return vMaintenance < maintenance ? -1 : 1;
		}
		return 0;
	}
	
	/**
		Find the first digits in a string and return those as the build number.
	*/
	private static int parseBuildNumber(String buildNumber)
	{
		int build = 0;
		Pattern p = Pattern.compile("(\\d+)");
		Matcher m = p.matcher(buildNumber);
		if (m.find())
		{
			build = Integer.parseInt(m.group());
		}
		return build;
	}

}
