/**
	$Id: MySQLEngine.java 6667 2014-12-18 12:14:15Z nicklas $

	Copyright (C) 2006, 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.dbengine;

import java.util.Collection;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import net.sf.basedb.core.hibernate.SchemaGenerator;

import org.hibernate.dialect.Dialect;
import org.hibernate.mapping.PersistentClass;

/**
	Database engine for My SQL.

	@author nicklas
	@version 2.0
	@base.modified $Date: 2014-12-18 13:14:15 +0100 (to, 18 dec 2014) $
*/
public class MySQLEngine
	extends AbstractDbEngine
	implements DbEngine
{
	/**
		Create MySQLEngine.
	*/
	public MySQLEngine()
	{}
	
	/*
		From the DbEngine interface
		-------------------------------------------
	*/
	/**
		Always false.
	*/
	@Override
	public boolean useThetaJoin()
	{
		return false;
	}
	
	/**
		Generates a catalog name by combining the catalog and
		schema name (catalog_schema).
		@since 3.4
	*/
	@Override
	public String getCatalogName(String catalog, String schema)
	{
		return catalog + "_" + schema;
	}
	
	/**
		Return null.
		@since 3.4
	*/
	@Override
	public String getSchemaName(String catalog, String schema)
	{
		return null;
	}
	
	/**
		Generate <code>CREATE SCHEMA &lt;catalog&gt;</code>.
		@since 3.4
	*/
	@Override
	public String getCreateSchemaSql(String catalog, String schema)
	{
		StringBuilder sql = new StringBuilder();
		sql.append("CREATE SCHEMA ").append(getQuotedName(catalog));
		return sql.toString();
	}

	
	/**
	 	Generate <code>ALTER TABLE &lt;catalog&gt;.&lt;table&gt; ADD [UNIQUE] 
	 	INDEX &lt;name&gt; (&lt;columns&gt;)</code>.
		The schema parameter is not supported.
	*/
	@Override
	public String getCreateIndexSql(String catalog, String schema, String table, String name, 
		Set<String> columns, boolean unique)
	{
		StringBuilder sql = new StringBuilder();
		sql.append("ALTER TABLE ");
		if (catalog != null) sql.append(getQuotedName(catalog)).append(".");
		sql.append(getQuotedName(table)).append(" ADD ");
		if (unique) sql.append("UNIQUE ");
		sql.append("INDEX ").append(getQuotedName(name)).append(" (");
		appendList(sql, columns);
		sql.append(")");
		return sql.toString();
	}
	
	/**
	 	Generate <code>DROP INDEX &lt;name&gt; ON &lt;catalog&gt;.&lt;table&gt;</code>.
		The schema parameter is not supported.
	*/
	@Override
	public String getDropIndexSql(String catalog, String schema, String table, String name, boolean unique)
	{
		StringBuilder sql = new StringBuilder();
		sql.append("DROP INDEX ").append(getQuotedName(name)).append(" ON ");
		if (catalog != null) sql.append(getQuotedName(catalog)).append(".");
		sql.append(getQuotedName(table));
		return sql.toString();
	}

	/**
		Generate <code>ALTER TABLE schema.table ADD CONSTRAINT name
		FOREIGN KEY (columns) REFERERENCES refTable (refColumns)</code>
		@since 3.1
	 */
	@Override
	public String getCreateForeignKeySql(String catalog, String schema,	String table, 
		String name, Set<String> columns, String refTable, Set<String> refColumns) 
	{
		StringBuilder sql = new StringBuilder();
		
		sql.append("ALTER TABLE ");
		if (schema != null) sql.append(getQuotedName(schema)).append(".");
		sql.append(getQuotedName(table));
		sql.append(" ADD CONSTRAINT ");
		sql.append(getQuotedName(name));
		sql.append(" FOREIGN KEY (");
		appendList(sql, columns);
		sql.append(") REFERENCES ");
		sql.append(getQuotedName(refTable));
		sql.append(" (");
		appendList(sql, refColumns);
		sql.append(")");
		return sql.toString();
	}
	
	/**
		Generate <code>ALTER TABLE &lt;catalog&gt;.&lt;table&gt; DROP FOREIGN KEY &lt;name&gt;</code>
		@since 3.0
	*/
	@Override
	public String getDropForeignKeySql(String catalog, String schema, String table, String name)
	{
		StringBuilder sql = new StringBuilder();
		sql.append("ALTER TABLE ");
		if (catalog != null) sql.append(getQuotedName(catalog)).append(".");
		sql.append(getQuotedName(table));
		sql.append(" DROP FOREIGN KEY ");
		sql.append(getQuotedName(name));
		return sql.toString();
	}
	
	/**
		Generate <code>ALTER TABLE catalog.table ADD PRIMARY KEY (columns)</code>
		@since 3.1
	*/
	@Override
	public String getCreatePrimaryKeySql(String catalog, String schema, String table, String name, Set<String> columns) 
	{
		StringBuilder sql = new StringBuilder();
		sql.append("ALTER TABLE ");
		if (catalog != null) sql.append(getQuotedName(catalog)).append(".");
		sql.append(getQuotedName(table));
		sql.append(" ADD PRIMARY KEY (");
		appendList(sql, columns);
		sql.append(")");
		return sql.toString();
	}

	/**
		Generate <code>ALTER TABLE catalog.table DROP PRIMARY KEY</code>
		@since 3.1
	*/
	@Override
	public String getDropPrimaryKey(String catalog, String schema, String table, String name) 
	{
		StringBuilder sql = new StringBuilder();
		sql.append("ALTER TABLE ");
		if (catalog != null) sql.append(getQuotedName(catalog)).append(".");
		sql.append(getQuotedName(table));
		sql.append(" DROP PRIMARY KEY");
		return sql.toString();
	}

	/**
		@return Always "null" (assuming that the ID column is an AUTO_INCREMENT column)
	*/
	@Override
	public String getInsertAutoIncrementSql(PersistentClass pc)
	{
		return "null";
	}

	/**
		Fix "alter table ... drop constraint ..." statements that are generated
		incorrectly by Hibernate.
		@since 3.4
	*/
	@Override
	public String inspectSchemaGenerationSQL(String sql, Dialect dialect,
			SchemaGenerator.Mode mode)
	{
		sql = super.inspectSchemaGenerationSQL(sql, dialect, mode);
		if (sql != null && sql.startsWith("alter table") && sql.contains("drop constraint"))
		{
			sql = sql.replace("drop constraint", "drop index");
		}
		return sql;
	}

	
	/**
		Return true.
	*/
	@Override
	public boolean analyzeAfterBatchOperation()
	{
		return true;
	}

	/**
		Generate <code>OPTIMIZE TABLE &lt;catalog&gt;.&lt;table&gt;</code>.
		The schema parameter is not supported.
	*/
	@Override
	public String getOptimizeTableSql(String catalog, String schema, String table)
	{
		catalog = catalog == null ? "" : getQuotedName(catalog) + ".";
		return "OPTIMIZE TABLE " + catalog + getQuotedName(table);
	}
	
	/**
		Generate <code>ANALYZE TABLE &lt;catalog&gt;.&lt;table&gt;</code>.
		The schema parameter is not supported.
	*/
	@Override
	public String getAnalyzeTableSql(String catalog, String schema, String table)
	{
		catalog = catalog == null ? "" : getQuotedName(catalog) + ".";
		return "ANALYZE TABLE " + catalog + getQuotedName(table);
	}

	/**
		Replace CREATE TABLE with CREATE TABLE IF NOT EXISTS...
		@since 2.6
	*/
	@Override
	public String makeSafeCreateTable(String sql, String catalog, String schema, String table)
	{
		Pattern p = Pattern.compile("CREATE TABLE", Pattern.CASE_INSENSITIVE);
		Matcher m = p.matcher(sql);
		sql = m.replaceFirst("CREATE TABLE IF NOT EXISTS");
		return sql;
	}
	
	/**
		Returns TRUE.
		@since 2.4
	*/
	@Override
	public boolean caseInsensitiveComparison()
	{
		return true;
	}
		
	
	/**
		Returns FALSE.
	*/
	@Override
	public boolean supportsColumnAliasInWhere()
	{
		return false;
	}
	/**
		Returns TRUE.
	*/
	@Override
	public boolean supportColumnAliasInOrderBy()
	{
		return true;
	}
	/**
		Returns TRUE.
	*/
	@Override
	public boolean supportColumnAliasInGroupBy()
	{
		return true;
	}
	/**
		Returns TRUE.
	*/
	@Override
	public boolean supportColumnAliasInHaving()
	{
		return true;
	}
	/**
		Return <code>LOG(&lt;value&gt;)</code>.
	*/
	@Override
	public String ln(String value)
	{
		return "LOG("+ value + ")";
	}
	/**
		Return <code>&lt;value&gt; RLIKE &lt;regexp&gt;</code>.
		@since 2.8
	*/
	@Override
	public String rlike(String value, String regexp)
	{
		return value + " RLIKE " + regexp;
	}
	
	/**
		Return <code>varchar(length) CHARACTER SET utf8 COLLATE utf8_bin</code>.
		@since 2.9
	 */
	@Override
	public String getCaseSensitiveVarchar(int length)
	{
		return "varchar(" + length + ") CHARACTER SET utf8 COLLATE utf8_bin";
	}
	/**
		Put quotes (`) around the name.
	*/
	@Override
	public String getQuotedName(String name)
	{
		return "`" + name + "`";
	}
	
	@Override
	public String getApproximateRowCountSql(String catalog, String schema, String table)
	{
		return "select TABLE_ROWS from INFORMATION_SCHEMA.TABLES where TABLE_NAME='" + table + "' and TABLE_SCHEMA='" + catalog + "'";
	}

	// -------------------------------------------

	private void appendList(StringBuilder sb, Collection<String> items)
	{
		int i = 0;
		for (String column : items)
		{
			if (i > 0) sb.append(", ");
			sb.append(getQuotedName(column));
			i++;
		}
	}

}
