/**
	$Id$

	Copyright (C) 2014 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.hibernate;

import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;

import org.hibernate.cfg.Configuration;
import org.hibernate.dialect.Dialect;
import org.hibernate.jdbc.Work;
import org.hibernate.mapping.Table;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.dbengine.DbEngine;
import net.sf.basedb.core.dbengine.TableInfo;
import net.sf.basedb.core.dbengine.TableInfo.ColumnInfo;
import net.sf.basedb.core.dbengine.TableInfo.ForeignKeyInfo;
import net.sf.basedb.core.dbengine.TableInfo.IndexInfo;

/**
	A work implementation that can create, update and drop indexes
	on tables.
	
	@author nicklas
	@since 3.4
*/
public class DbIndexWork 
	implements Work
{

	private static final org.slf4j.Logger log = 
			org.slf4j.LoggerFactory.getLogger(DbIndexWork.class);

	
	private final Configuration cfg;
	private final Dialect dialect;
	private final DbEngine dbEngine;
	private final boolean isVerbose;
	private final boolean silent;
	private final boolean dropIndexes;
	private final boolean updateIndexes;
	
	public DbIndexWork(Configuration cfg, Dialect dialect, DbEngine dbEngine, 
		boolean verbose, boolean silent, 
		boolean dropIndexes, boolean updateIndexes)
	{
		this.cfg = cfg;
		this.dialect = dialect;
		this.dbEngine = dbEngine;
		this.isVerbose = verbose && !silent;
		this.silent = silent;
		this.dropIndexes = dropIndexes;
		this.updateIndexes = updateIndexes;
	}
	
	@Override
	public void execute(Connection connection) 
		throws SQLException
	{
		Iterator<Table> tables = (Iterator<Table>)cfg.getTableMappings();
		String defaultCatalog = cfg.getProperty("hibernate.default_catalog");
		String defaultSchema = cfg.getProperty("hibernate.default_schema");
		
		DatabaseMetaData metaData = connection.getMetaData();
		while (tables.hasNext())
		{
			Table table = tables.next();
			
			String catalog = table.getCatalog() == null ? defaultCatalog : table.getCatalog();
			String schema = table.getSchema() == null ? defaultSchema : table.getSchema();
			
			if (!silent)
			{
				System.out.println("=================");
				System.out.println("Table   : " + table.getName());
			}
			if (isVerbose)
			{
				System.out.println("Catalog : " + table.getCatalog() + " (" + defaultCatalog + ")");
				System.out.println("Schema  : " + table.getSchema()+ " (" + defaultSchema + ")");
			}
	
			if (isVerbose)
			{
				System.out.println("Database information");
				System.out.println("--------------------");
			}
			
			TableInfo tiDb = null;
			try
			{
				tiDb = new TableInfo(table, metaData, defaultCatalog, defaultSchema);
			}
			catch (SQLException ex)
			{
				throw new BaseException(ex);
			}
	
			if (isVerbose && tiDb != null)
			{
				// Write all columns
				for (ColumnInfo ci : tiDb.getColumns())
				{
					System.out.println("  Column       : " + ci);
				}
				
				// Write primary key
				System.out.println("  Primary key  : " + tiDb.getPrimaryKey());
			
				// Write foreign keys
				for (ForeignKeyInfo fk : tiDb.getForeignKeys())
				{
					System.out.println("  Foreign key  : " + fk);
				}
			
				// Write indexes and unique constraints
				for (IndexInfo ii : tiDb.getIndexes())
				{
					if (ii.isUnique())
					{
						System.out.println("  Unique       : " + ii);
					}
					else
					{
						System.out.println("  Index        : " + ii);
					}
				}
			}
			
			if (isVerbose)
			{
				System.out.println("Hibernate information");
				System.out.println("---------------------");
			}
			
			TableInfo tiHib = new TableInfo(table, dialect);
			if (isVerbose)
			{
				// Write all columns
				for (ColumnInfo ci : tiHib.getColumns())
				{
					System.out.println("  Column       : " + ci);
				}
				
				// Write primary key
				System.out.println("  Primary key  : " + tiHib.getPrimaryKey());
			}
			
			// Write foreign keys
			for (ForeignKeyInfo fk : tiHib.getForeignKeys())
			{
				if (!silent) 
				{
					System.out.println("  Foreign key  : " + fk);
				}
				
				// Check that an index for the foreign key exists
				String indexName = tiDb.findIndexName(fk.getName(), fk.getFkColumns());
				boolean indexExists = indexName != null;
				String createSql = 
					dbEngine.getCreateIndexSql(catalog, schema,
						table.getName(), fk.getName(), fk.getFkColumns(), false);
				boolean actionCreate = updateIndexes && !indexExists;
				
				if (!silent)
				{
					System.out.println("    Indexed    : " + indexExists + (indexExists ? "(" + indexName + ")" : ""));
					System.out.println("    INDEX-SQL  : " + createSql);
					System.out.println("    Actions    : " + (actionCreate ? "CREATE INDEX" : ""));
				}
				if (actionCreate)
				{
					Statement st = null;
					try
					{
						st = connection.createStatement();
						if (actionCreate) 
						{
							log.info("Creating foreign key index: + " + createSql);
							st.executeUpdate(createSql);
						}
						connection.commit();
						st.close();
						st = null;
					}
					catch (SQLException ex)
					{
						log.error("Exception", ex);
						throw new BaseException(ex);
					}
					finally
					{
						if (st != null)
						{
							try
							{
								st.close();
							}
							catch (Throwable t)
							{
								log.error("Exception", t);
							}
						}
					}
				}
			}
	
			// Write indexes and unique constraints
			for (IndexInfo ii : tiHib.getIndexes())
			{
				if (!silent)
				{
					if (ii.isUnique())
					{
						System.out.println("  Unique       : " + ii);
					}
					else
					{
						System.out.println("  Index        : " + ii);
					}
				}
				
				List<IndexInfo> existingIndexes = tiDb.findAll(ii.getColumns());
				Set<String> duplicateNames = new TreeSet<String>(String.CASE_INSENSITIVE_ORDER);
				//String dbName = tiDb.findIndexName(ii.getName(), ii.getColumns());
				boolean exists = existingIndexes.size() > 0;
				boolean duplicates = existingIndexes.size() > 1;
				boolean safeToDrop = exists && tiDb.safeToDrop(ii);
	
				String dbName = exists ? existingIndexes.get(0).getName() : null;
				if (duplicates)
				{
					// If there are duplicates we find the index with the
					// same name as defined by Hibernate or with the shortest name
					// and claim that as the 'true' index. The others will be dropped
					String shortestName = null;
					for (IndexInfo dup : existingIndexes)
					{
						String name = dup.getName();
						duplicateNames.add(name);
						if (shortestName == null || name.length() < shortestName.length())
						{
							shortestName = name;
						}
					}
					
					dbName = duplicateNames.contains(ii.getName()) ? ii.getName() : shortestName;
					duplicateNames.remove(dbName);
				}
				
				String dropSql = exists ? 
					dbEngine.getDropIndexSql(catalog, schema, 
						table.getName(), dbName, ii.isUnique()) : "";
				String createSql = 
					dbEngine.getCreateIndexSql(catalog, schema,
						table.getName(), dbName == null ? table.getName() + "_" + ii.getName() : dbName, 
					ii.getColumns(), ii.isUnique());
				boolean actionDrop = dropIndexes && exists && safeToDrop;
				boolean actionCreate = updateIndexes && (actionDrop || !exists);
				boolean actionDropDuplicates = (dropIndexes || updateIndexes) && duplicates;
				
				if (!silent)
				{
					System.out.println("    Exists     : " + exists + (exists ? "(" + dbName + ")" : ""));
					System.out.println("    Duplicates : " + duplicates + (duplicates ? "(" + duplicateNames + ")" : ""));
					System.out.println("    Safe drop  : " + safeToDrop);
					System.out.println("    DROP-SQL   : " + dropSql);
					System.out.println("    CREATE-SQL : " + createSql);
					System.out.println("    Actions    : " + (actionDrop ? "DROP " : "") + 
						(actionCreate ? "CREATE" : "") + 
						(actionDropDuplicates ? "DROP-DUPLICATES" : ""));
				}
				if (actionDrop || actionCreate || actionDropDuplicates)
				{
					Statement st = null;
					try
					{
						st = connection.createStatement();
						if (actionDropDuplicates)
						{
							for (String dupName : duplicateNames)
							{
								String dropDupSql = dbEngine.getDropIndexSql(catalog, schema, 
										table.getName(), dupName, ii.isUnique());
								log.info("Dropping duplicate index: " + dropDupSql);
								st.executeUpdate(dropDupSql);
							}
						}
						if (actionDrop) 
						{
							log.info("Dropping index: + " + dropSql);
							st.executeUpdate(dropSql);
						}
						if (actionCreate) 
						{
							log.info("Creating index: + " + createSql);
							st.executeUpdate(createSql);
						}
						connection.commit();
						st.close();
						st = null;
					}
					catch (SQLException ex)
					{
						log.error("Exception", ex);
						throw new BaseException(ex);
					}
					finally
					{
						if (st != null)
						{
							try
							{
								st.close();
							}
							catch (Throwable t)
							{
								log.error("Exception", t);
							}
						}
					}
				}
				
			}

			if (!silent)
			{
				System.out.println("=================");
				System.out.println("");
			}
		}
		
	}
	

}
