/**
	$Id$

	Copyright (C) 2014 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.hibernate;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.ServiceSessionControl;

import org.hibernate.cfg.Configuration;

/**
	Class for holding configuration options for creating
	extra session factories for services. Used by 
	{@link ServiceSessionControl#buildSessionFactory(SessionFactoryConfiguration)}.

	@author nicklas
	@since 3.4
*/
public class SessionFactoryConfiguration 
{

	private final String schema;
	private final ClassLoader loader;
	private final Configuration cfg;
	
	private String appId;
	private int schemaVersion;
	private int installedSchemaVersion;
	private boolean useSeparateConnections;
	
	/**
		Create a configuration instance. A schema is required and is the
		place were tabales for configured classes will be created.
		
		A class loader is needed so that Hibernate can load the classes from
		extension classes.
		
		It is recommended that an "application id" and schema version is set
		by calling {@link #setSchemaVersion(String, int)}. Then, BASE can
		check the currently installed schema version and only update the
		database if there is a change. Note that the schema version must
		always increase from one version to the next.
		
		Use {@link #getHibernateConfig()} to retrieve the backend Hibernate
		configuration instance. If you have annotated entity classes you 
		can add them with {@link Configuration#addAnnotatedClass(Class)}, but
		there are a lots of other configuration possibilities that are beyond
		the scope of this documentation. A good Hibernate tutorial is recommended
		(eg. http://hibernate.org/orm/documentation/)
		
		Also note that by default the configured session factory need to be
		"piggy-backed" on a BASE {@link DbControl} transaction. Create
		Hibernate sessions with {@link ServiceSessionControl#newSession(org.hibernate.SessionBuilder, DbControl)}.
		
		If you want to use connections/transactions that are independent from BASE 
		call {@link #useSeparateConnections()} but make sure you also get connection
		pooling configured. You can still use "piggy-backed" 
		connections.

		@param schema The schema name the extension wants to use for it's database tables
		@param classLoader The class loader that can load classes for the extension (needed by Hibernate)
	*/
	public SessionFactoryConfiguration(String schema, ClassLoader loader)
	{
		this.schema = schema;
		this.loader = loader;
		this.cfg = new Configuration();
	}
	
	/**
		Get the schema.
	*/
	public String getSchema()
	{
		return schema;
	}
	
	/**
		Get the schema.
	*/
	public ClassLoader getClassLoader()
	{
		return loader;
	}
	
	/**
		Get the underlying Hibernate configuration instance. You will need
		this to defined mappings between entities and database tables.
	*/
	public Configuration getHibernateConfig()
	{
		return cfg;
	}
	
	/**
		Set the schema version of the current configuration.
		If this information has been set, BASE will check the
		current schema version registration and only issue a
		database update in case the registered schema version
		is lower than the given. If the registered version is
		higher the session factory creation fails.
		
		If no schema version has been set, BASE will always try a
		database update.
		
		@param appId The application ID
		@param schemaVersion The schema version to install
	*/
	public void setSchemaVersion(String appId, int schemaVersion)
	{
		this.appId = appId;
		this.schemaVersion = schemaVersion;
	}
	
	/**
		Get the application id for schema versioning.
	*/
	public String getAppId()
	{
		return appId;
	}
	
	/**
		Get the schema version of the current configuration.
	*/
	public int getSchemaVersion()
	{
		return schemaVersion;
	}
	
	/**
		Get the schema version that was installed prior to updating it.
		Note that this value is only available after calling
		{@link ServiceSessionControl#buildSessionFactory(SessionFactoryConfiguration)}.
	*/
	public int getInstalledSchemaVersion()
	{
		return installedSchemaVersion;
	}
	
	/**
		Callback intended for BASE core to set the installed schema version.
	*/
	public void setInstalledSchemaVersion(int schemaVersion)
	{
		this.installedSchemaVersion = schemaVersion;
	}
	
	/**
		Should the session factory be configured to use separate
		connections or is piggy-backed sessions required?
	*/
	public boolean useSeparateConnections()
	{
		return useSeparateConnections;
	}
	
	/**
		Set to TRUE to use connections separate from BASE connections. 
		If you enable this, you will probably also want to configure 
		connection pooling since the default connection pooling is
		not recommended for production use. Connection pooling
		is configured on {@link #getHibernateConfig()} object. Basically
		you need to set the 'connection.provider_class' property 
		to 'org.hibernate.connection.C3P0ConnectionProvider' and
		some other properties. You'll need to look for this
		documentation in other places.
	*/
	public void setUseSeparateConnections(boolean separate)
	{
		this.useSeparateConnections = separate;
	}
}
