/**
	$Id$

	Copyright (C) 2014 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.snapshot;

import java.io.Serializable;
import java.util.List;

import net.sf.basedb.core.AnnotationType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.Unit;
import net.sf.basedb.util.filter.Filter;
import net.sf.basedb.util.units.UnitConverter;

/**
	Utility implementation for simple loading of primary annotations
	via {@link SnapshotManager}.

	@author Nicklas
	@since 3.4
*/
public class AnnotationLoaderUtil
	implements Filter<AnnotationSnapshot>
{
	private final DbControl dc;
	private final SnapshotManager manager;
	private final AnnotationType at;
	private final int atId;
	private final Type valueType;
	private final Unit defultUnit;
	
	private AnnotationSnapshot snapshot;
	private UnitConverter converter;
	private Unit unit;
	private String unitSymbol;
	
	/**
		Create a loder that uses the given snapshot manager to load annotations
		for a single annotation type.
	*/
	public AnnotationLoaderUtil(DbControl dc, SnapshotManager manager, AnnotationType at)
	{
		this.dc = dc;
		this.manager = manager;
		this.at = at;
		this.atId = at.getId();
		this.valueType = at.getValueType();
		this.defultUnit = at.getDefaultUnit();
	}
	
	/*
		From the Filter interface
		-------------------------
	*/
	@Override
	public boolean evaluate(AnnotationSnapshot snapshot)
	{
		return snapshot.getAnnotationTypeId() == atId;
	}
	// -------------------------

	/**
		The ID of the annotation type that this loader load annotations from.
	*/
	public int getId()
	{
		return atId;
	}
	
	/**
		The annotation type that this loader load annotations from.
	*/
	public AnnotationType getAnnotationType()
	{
		return at;
	}
	
	/**
		Find and load annotations from the given annotation set snapshot. The
		loader will only load primary annotations.
		@return TRUE if an annotation was found, FALSE if not
	*/
	public boolean find(AnnotationSetSnapshot setSnapshot)
	{
		List<AnnotationSnapshot> list = manager.findAnnotations(dc, setSnapshot, this, false);
		if (list.size() > 0)
		{
			snapshot = list.get(0);
			unit = snapshot.getUnit(dc);
		}
		else
		{
			snapshot = null;
			unit = null;
		}
		if (unit != null)
		{
			converter = unit.getUnitConverter(defultUnit);
			unitSymbol = unit.getDisplaySymbol();
		}
		else
		{
			converter = null;
			unitSymbol = null;
		}
		return snapshot != null;
	}
	
	/**
		Get the snapshot for the last annotation found when calling
		{@link #find(AnnotationSetSnapshot)}
		@return A snapshot or null if no annotations are found
	*/
	public AnnotationSnapshot getSnapshot()
	{
		return snapshot;
	}
	
	/**
		Get the values from the last annotation found when calling
		{@link #find(AnnotationSetSnapshot)}
		@return A list of annotation values or null if no annotations are found
	*/
	public List<? extends Serializable> getValues()
	{
		return snapshot.getValues(converter, valueType);
	}
	
	/**
		Get the unit of the last annotation values that was found
		by {@link #find(AnnotationSetSnapshot)}.
		@return A string or null if the annotation doesn't use units
	*/
	public Unit getUnit()
	{
		return unit;
	}

	
	/**
		Get the unit symbol of the last annotation values that was found
		by {@link #find(AnnotationSetSnapshot)}.
		@return A string or null if the annotation doesn't use units
	*/
	public String getUnitSymbol()
	{
		return unitSymbol;
	}

	
}
