/**
	$Id: AssayPropertyField.java 5373 2010-08-03 09:11:54Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.export.spotdata;

import java.util.Collection;
import java.util.Collections;

import net.sf.basedb.core.BioAssay;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Metadata;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.Metadata.PropertyPath;
import net.sf.basedb.util.EqualsHelper;
import net.sf.basedb.util.formatter.Formatter;

/**
	Assay field implementation that exports a named property 
	from a bioassay. The default setting will export the "name"
	property. One instance of this class is considered to
	be equal to another if they are configured with the same property
	and title.

	@author Nicklas
	@since 2.15
	@base.modified $Date: 2010-08-03 11:11:54 +0200 (ti, 03 aug 2010) $
*/
public class AssayPropertyField
	implements AssayField
{
	private final Metadata<BioAssay> metadata;
	private PropertyPath<BioAssay, ?> propertyPath;
	private String property;
	private String title;
	private Type type;
	private Formatter<?> formatter;
	
	/**
		Create a new annotation field. The property is automatically
		set to "name". Call {@link #setProperty(String)}
		and {@link #setTitle(String)} to change.
		@see ExportableFieldFactory#assayProperty(String, String, Type, Formatter)
	*/
	public AssayPropertyField()
	{
		this.metadata = Metadata.getInstance(BioAssay.class);
		setProperty("name");
		setTitle("Name");
		setType(Type.STRING);
	}
	
	/*
		From the AssayField interface
		-------------------------------
	*/
	@Override
	public String getTitle()
	{
		return title == null ? "" : title;
	}
	/**
		Return all values that are given by the specified property.
	*/
	@Override
	public Collection<?> getAssayValue(DbControl dc, BioAssay ba)
	{
		Collection c = null;
		Object value = propertyPath.getValue(dc, ba);
		if (value != null)
		{
			c = value instanceof Collection ? 
					(Collection)value : Collections.singleton(value);
		}
		return c;
	}
	@Override
	public Type getType()
	{
		return type;
	}
	@Override
	public Formatter<?> getFormatter()
	{
		return formatter;
	}
	/**
		@return Always FALSE
	*/
	@Override
	public boolean isAnnotation()
	{
		return false;
	}
	// -----------------------------------

	/**
		An assay field is equal to another assay field if it uses the same
		property and has the same title.
	*/
	@Override
	public boolean equals(Object obj)
	{
		if (this == obj) return true;
		if (!(obj instanceof AssayPropertyField)) return false;
		AssayPropertyField other = (AssayPropertyField)obj;
		return EqualsHelper.equals(this.getProperty(), other.getProperty()) 
			&& EqualsHelper.equals(this.getTitle(), other.getTitle());
	}

	@Override
	public int hashCode()
	{
		return 13 * EqualsHelper.hashCode(getProperty()) + 
			EqualsHelper.hashCode(getTitle());
	}

	@Override
	public String toString()
	{
		return "AssayPropertyField(title=" + getTitle() + "; property=" + getProperty() + ")";
	}

	/**
		Set the property that should be exported. For syntax
		see {@link Metadata#getPropertyPath(String, boolean)}.
		Do not forget to also change the title.
	*/
	public void setProperty(String property)
	{
		this.property = property;
		this.propertyPath = metadata.getPropertyPath(property, true);
	}
	
	/**
		Get the configured property.
	*/
	public String getProperty()
	{
		return property;
	}
	
	/**
		Set the title of the exported property.
	*/
	public void setTitle(String title)
	{
		this.title = title;
	}

	/**
		Set the data type of the exported property.
		@since 2.15
	*/
	public void setType(Type type)
	{
		this.type = type;
	}
	
	/**
		Set a formatter to use for formatting the exported values.
	*/
	public void setFormatter(Formatter<?> formatter)
	{
		this.formatter = formatter;
	}
	
}
