/**
	$Id: SerialBfsExporter.java 5319 2010-04-20 08:02:16Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.export.spotdata;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import net.sf.basedb.core.BioAssay;
import net.sf.basedb.core.Path;
import net.sf.basedb.util.bfs.DataWriter;

/**
	BFS exporter implementation that exports spot data in serial format. This
	format creates one file per assay in the bioassay set. Each data file contains
	columns for the registered spot fields. For example, the 'sdata1'
	file contains channel 1 and 2 intensity for the first assay, 
	and 'sdata2' contains the same data for the second assay, etc.

	@author Nicklas
	@version 2.15
	@base.modified $Date: 2010-04-20 10:02:16 +0200 (ti, 20 apr 2010) $
*/
public class SerialBfsExporter
	extends BfsExporter
{
	
	public static final String SUBTYPE = "serial";
	
	/**
		Create a new serial BFS exporter.
	*/
	public SerialBfsExporter()
	{}

	
	/*
		From the BfsExporter class
		--------------------------
	*/
	/**
		@return "serial"
	*/
	@Override
	public String getBFSSubtype()
	{
		return SUBTYPE;
	}
	
	/**
		Create one data writer for each assay in the bioassay set. Each data 
		writer is configured with one column per spot field.
	*/
	@Override
	protected List<DataWriter> createDataWriters()
		throws IOException
	{
		prepareAssayIndexMap(getBioAssays(), 0, 1);
		List<BioAssay> bioassays = getBioAssays();
		int numSpotFields = getSpotFields().size();
		List<DataWriter> writers = new ArrayList<DataWriter>();
		for (BioAssay ba : bioassays)
		{
			DataWriter writer = createDataWriter(ba, Path.makeSafeFilename(ba.getName() + "-" + ba.getId() + ".txt", ""));
			writer.setColumnCount(numSpotFields);
			writers.add(writer);
		}
		return writers;
	}

	/**
		The column number is used to find the assay index. Then,
		the spot data is copied to the correct data writer.
	*/
	@Override
	protected void arrangeData(int position, short column, 
		 Object[][] data, Object[] spotData)
	{
		int columnToArrayIndex = getAssayIndex(column);
		Object[] assayData = data[columnToArrayIndex];
		for (int i = 0; i < assayData.length; ++i)
		{
			assayData[i] = spotData[i];
		}
	}
	// -------------------------------------------------------
	
}
