/**
	$Id$
	
	Copyright (C) 2012 Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/
	
	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.
	
	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.loader;

import net.sf.basedb.core.BioMaterial;
import net.sf.basedb.core.BioMaterialList;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ItemResultIterator;
import net.sf.basedb.util.overview.Node;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.node.BioMaterialListMemberNameGenerator;
import net.sf.basedb.util.overview.node.ChildNodeDirection;
import net.sf.basedb.util.overview.node.NameableNameGenerator;
import net.sf.basedb.util.overview.node.NodeFactory;
import net.sf.basedb.util.overview.validator.NodeValidator;


/**
	Node loader implementation for biomaterial lists. Biomaterial
	lists can only be loaded as root nodes. The forward-loading
	direction goes to the biomaterial that is member of the list.
	
	@author nicklas
	@since 3.2
*/
public class BioMaterialListLoader 
	extends BasicItemNodeLoader<BioMaterialList> 
{
	public BioMaterialListLoader()
	{
		super(Item.BIOMATERIALLIST, ALLOW_ROOT_NODE, 
				new NameableNameGenerator<BioMaterialList>("list", "List"));
	}
	
	/*
		From the AbstractNodeLoader class
		----------------------------------
	*/	
	/**
		Loads property nodes of a biomaterial list. 
		<ul>
		<li>Annotations: {@link AnnotationLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		<li>Any-to-any: {@link AnyToAnyLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		</ul>
	*/
	@Override
	protected void loadPropertyChildNodes(DbControl dc, OverviewContext context, Node listNode)
	{
		getNodeLoader(context, Item.ANNOTATION).createPropertyNode(dc, context, listNode);
		getNodeLoader(context, Item.ANYTOANY).createPropertyNode(dc, context, listNode);
	}

	/**
		Loads the list members.
	*/
	@Override
	protected void loadForwardChildNodes(DbControl dc, OverviewContext context, Node listNode)
	{
		BioMaterialList list = (BioMaterialList)listNode.getItem(dc);
		getNodeLoader(context, list.getMemberType()).createForwardNode(dc, context, listNode);
	}
	// ---------------------------------------

	/**
		Utiltity method to be used by biomaterial loaders to load members of a biomaterial list.
		The loader should be one of {@link SampleLoader}, {@link ExtractLoader} or
		{@link BioSourceLoader}. The member nodes are created as child nodes
		to the list node without any intermediate folder node.
		
		@param loader The load implementation
		@param listNode The parent biomaterial list node
	*/
	@SuppressWarnings("unchecked")
	public static <T extends BioMaterial> void loadMemberNodes(BasicItemNodeLoader<T> loader, DbControl dc, OverviewContext context, Node listNode)
	{
		NodeValidator<T> validator = loader.getNodeValidator(context);
		NodeFactory<T> factory = loader.getNodeFactory(dc, context, validator, new BioMaterialListMemberNameGenerator());
		BioMaterialList list = (BioMaterialList)listNode.getItem(dc);
		
		// Find all members of the list
		ItemQuery<T> query = (ItemQuery<T>)context.initQuery(list.getMemberBioMaterials(), "name");
		
		ItemResultIterator<T> it = query.iterate(dc);
		while (it.hasNext())
		{
			T bioMaterial = it.next();
			loader.createItemNode(factory, bioMaterial, bioMaterial, false, listNode, ChildNodeDirection.ALL);
		}
	}

}
