/**
	$Id: ArrayDesignValidator.java 5814 2011-10-18 11:46:59Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.validator;

import java.util.HashSet;
import java.util.Set;

import net.sf.basedb.core.ArrayDesign;
import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.DerivedBioAssay;
import net.sf.basedb.core.Experiment;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.Project;
import net.sf.basedb.core.RawBioAssay;
import net.sf.basedb.util.overview.Fix;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.Validator;
import net.sf.basedb.util.overview.filter.ItemTypeFilter;
import net.sf.basedb.util.overview.Node;

/**
	Validator implementation for array designs. Validation rules:
	<ul>
	<li>Missing item: {@link Validator#MISSING_ARRAYDESIGN}
	<li>Access denied: {@link Validator#DENIED_ARRAYDESIGN}
	<li>Non-default array design: {@link Validator#NONDEFAULT_ARRAYDESIGN}
	<li>Non-matching array design: {@link Validator#NONMATHCING_ARRAYDESIGN}
	<li>Use of multiple array design in an experiment: {@link Validator#MULTIPLE_ARRAYDESIGNS}
	</ul>

	@author Nicklas
	@version 2.10
	@base.modified $Date: 2011-10-18 13:46:59 +0200 (ti, 18 okt 2011) $
*/
public class ArrayDesignValidator
	extends NameableNodeValidator<ArrayDesign>
{
	
	public ArrayDesignValidator()
	{
		super(Validator.MISSING_ARRAYDESIGN, Validator.DENIED_ARRAYDESIGN);
	}
	
	/* 
		From BasicValidator class
		-------------------------
	*/
	@Override
	public void postValidate(DbControl dc, OverviewContext context, Node node, Node parentNode)
	{
		super.postValidate(dc, context, node, parentNode);
		Project project = context.getProject();
		ArrayDesign design = (ArrayDesign)node.getItem();
		if (project != null)
		{
			if (!project.isDefaultItem(design) && project.hasDefaultForItem(design))
			{
				context.createFailure(Validator.NONDEFAULT_ARRAYDESIGN, node, null,
					new Fix("Change default array design for project", project)
				);
			}
		}
		
		// Check that a raw bioassay has same "direct" array desing as
		// via physical bioassay->slide...
		if (parentNode != null && parentNode.getItemType() == Item.RAWBIOASSAY)
		{
			RawBioAssay rba = (RawBioAssay)parentNode.getItem(dc);
			try
			{
				DerivedBioAssay bioAssay = rba.getParentBioAssay();
				if (bioAssay != null)
				{
					ArrayDesign other = bioAssay.getArrayDesign();
					if (other != null && !design.equals(other))
					{
						context.createFailure(Validator.NONMATHCING_ARRAYDESIGN, parentNode, null);
					}
				}
			}
			catch (PermissionDeniedException ex)
			{}
		}
		
		// Check if experiment is using more than one array design
		checkExperimentArrayDesign(dc, context, parentNode, design);
	}

	@Override
	public boolean preMissingItem(DbControl dc, OverviewContext context, Node parentNode)
	{
		checkExperimentArrayDesign(dc, context, parentNode, null);
		return super.preMissingItem(dc, context, parentNode);
	}

	/**
		The suggested fix is to add an array design to the
		raw bioassay.
	*/
	@Override
	protected Fix getMissingItemFix(DbControl dc, Node parentNode)
	{
		Fix fix = null;
		Item parentType = parentNode.getItemType();
		BasicItem parentItem = parentNode.getItem();
		if (parentType == Item.RAWBIOASSAY)
		{
			fix =  new Fix("Add array design to raw bioassay", parentItem);
		}
		return fix;
	}
	// ----------------------------

	/**
		Check if an experiment has used more than one array design.
		The context cache keep a Set with all arraydesign we have 
		seen so for for a given experiment (including null)
	*/
	@SuppressWarnings("unchecked")
	private void checkExperimentArrayDesign(DbControl dc, OverviewContext context, Node parentNode, ArrayDesign design)
	{
		// We should only check array design that are connected to raw bioassays
		if (parentNode == null || parentNode.getItemType() != Item.RAWBIOASSAY) return;
		
		Node experimentNode = parentNode.getFirstParent(new ItemTypeFilter(Item.EXPERIMENT));
		if (experimentNode != null)
		{
			Experiment exp = (Experiment)experimentNode.getItem(dc);
			String cacheKey = "experiment." + exp.getId() + ".arraydesign";
			Set<ArrayDesign> used = (Set<ArrayDesign>)context.getCachedObject(cacheKey);
			if (used == null)
			{
				used = new HashSet<ArrayDesign>();
				context.setCachedObject(cacheKey, used);
			}
			if (used.add(design))
			{
				if (used.size() == 2)
				{
					context.createFailure(Validator.MULTIPLE_ARRAYDESIGNS, experimentNode, null);
				}
			}
		}
	}
}
