/**
	$Id: DefaultItemFunction.java 6127 2012-09-14 07:28:35Z nicklas $

	Copyright (C) 2011 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.parser;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Stack;

import org.nfunk.jep.ParseException;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemSubtype;
import net.sf.basedb.core.Nameable;
import net.sf.basedb.core.Project;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.jep.JepFunction;

/**
	JEP function that can be used to return the name of a default item in
	a project. The constructor need to specify the project and the type
	of item to look for. Use 'default()' or 'default(col("header"))' in the
	JEP expression to use this function. The first variant will always
	return the default item if one exists for the project. The second
	variant will return the default item only if the 'col("header")' part
	is null or an empty string.
	<p>
	Before this function is used the {@link #setCurrentSubtype(ItemSubtype)}
	can be called to limit the search to items with a specific subtype. If
	there is more than one matching item in a project, it is undefined which
	one is returned, but multiple uses of this function will return the same 
	item.

	@author Nicklas
	@since 3.1
	@base.modified $Date: 2012-09-14 09:28:35 +0200 (fr, 14 sep 2012) $
*/
public class DefaultItemFunction<T extends Nameable>
	implements JepFunction
{

	private final DbControl dc;
	private final Project project;
	private final Item itemType;
	private ItemSubtype subtype;
	private int numParameters;
	private Map<Object, T> itemCache;
	
	/**
		Create a new function instance.
		@param dc The DbControl to use when loading items from the database
		@param project The project (null is allowed)
		@param itemType The type of item to look for
	*/
	public DefaultItemFunction(DbControl dc, Project project, Item itemType)
	{
		this.dc = dc;
		this.project = project;
		this.itemType = itemType;
		this.itemCache = new HashMap<Object, T>();
	}
	
	
	/*
		From the JepFunction interface
		-------------------------------------------
	*/
	/**
		@return The string "default"
	*/
	@Override
	public String getFunctionName()
	{
		return "default";
	}
	// -------------------------------------------

	/*
		From the PostfixMathCommandI interface
		-------------------------------------------
	*/

	@Override
	public int getNumberOfParameters()
	{
		return -1; // We accept 0 or 1 parameter
	}
	
	@Override
	public void setCurNumberOfParameters(int n)
	{
		this.numParameters = n;
	}

	@Override
	public boolean checkNumberOfParameters(int n)
	{
		return n == 0 || n == 1;
	}

	@SuppressWarnings("unchecked")
	@Override
	public void run(Stack stack) 
		throws ParseException
	{
		if (stack == null) 
		{
			throw new ParseException("Stack is null");
		}

		Object param = numParameters == 1 ? stack.pop() : null;
		// If there is no parameter or if it is empty...
		if (param == null || Values.getStringOrNull(param.toString()) == null)
		{
			// Find the default item and get the name...
			T item = findDefaultItem();
			param = item == null ? null : item.getName();
		}
		// Finally, put the result back on the stack
		stack.push(param);
	}
	// -------------------------------------------
	
	/**
		Find a default item. If a subtype has been set
		with {@link #setCurrentSubtype(ItemSubtype)} the
		{@link Project#findDefaultItems(DbControl, ItemSubtype, boolean)} 
		method is used in non-strict mode, otherwise the 
		{@link Project#findDefaultItems(DbControl, Item, boolean)}
		is used in strict mode.
		@return A matching default item or null if none could be found
	*/
	@SuppressWarnings("unchecked")
	public T findDefaultItem()
	{
		if (project == null) return null;
		T item = null;
		if (itemCache.containsKey(subtype))
		{
			item = itemCache.get(subtype);
		}
		else
		{
			List<T> items = subtype == null ? 
				(List<T>)project.findDefaultItems(dc, itemType, true) :
				(List<T>)project.findDefaultItems(dc, subtype, false);
			item = items.size() > 0 ? items.get(0) : null;
			itemCache.put(subtype, item);
		}
		return item;
	}
	
	/**
		Set the current subtype to use when looking for default items.
	*/
	public void setCurrentSubtype(ItemSubtype subtype)
	{
		this.subtype = subtype;
	}
}
