/**
	$Id: ConnectionManagerUtil.java 6520 2014-08-18 11:29:37Z nicklas $

	Copyright (C) 2011 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.uri;

import java.net.URI;
import java.util.Iterator;
import java.util.List;

import org.apache.commons.collections.IteratorUtils;

import net.sf.basedb.core.Application;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.File;
import net.sf.basedb.core.ItemNotFoundException;
import net.sf.basedb.util.extensions.ClientContext;
import net.sf.basedb.util.extensions.ExtensionsFilter;
import net.sf.basedb.util.extensions.ExtensionsInvoker;
import net.sf.basedb.util.extensions.Registry;
import net.sf.basedb.util.extensions.SingleExtensionFilter;
import net.sf.basedb.util.extensions.manager.ExtensionsManager;

/**
	Collects utility methods related to connection managers.
	Connection manager factories should be registered as 
	extensions to the {@link #EXTENSION_POINT_ID net.sf.basedb.core.uri.connection-manager}
	extension point. The methods in this class can then be used to
	get information about installed connection manager factories and
	to use them for retreiving external file data and information.

	@author Nicklas
	@since 3.0
	@base.modified $Date: 2014-08-18 13:29:37 +0200 (må, 18 aug 2014) $
*/
public class ConnectionManagerUtil
{
	
	/**
		The ID of the connection manager extension point.
	*/
	public static final String EXTENSION_POINT_ID = "net.sf.basedb.core.uri.connection-manager";
	
	/**
		Get a list with all connection manager factories. No context
		information will be available to the action factories that
		creates the connection manager factories.
		
		@return A list with the factories
	*/
	@SuppressWarnings("unchecked")
	public static List<ConnectionManagerFactory> getFactories()
	{
		ExtensionsInvoker<ConnectionManagerFactory> invoker = getInvoker(null, null);
		List list = IteratorUtils.toList(invoker.iterator(), Math.max(invoker.getNumExtensions(), 1));
		return list;
	}
	
	/**
	 	Get the connection manager factory with the given extension id.
	 	No context information will be available to the factory
	 	when it is created.
	 	
		@param id The ID of the factory
		@return A factory instance or null if not found
	*/
	public static ConnectionManagerFactory getFactory(String id)
	{
		if (id == null) return null;
		ExtensionsInvoker<ConnectionManagerFactory> invoker = getInvoker(null, new SingleExtensionFilter(id));
		ConnectionManagerFactory factory = getFirst(invoker);
		return factory;
	}
		
	/**
		Find a factory for the given URI. This method will iterate
		the registered factories in their priority/index order and 
		call {@link ConnectionManagerFactory#supports(URI)}
		for all factories that supports auto-detection. The first
		one that gives a successful response is returned. If no factory
		supports the given URI, null is returned.
		
		<p>
		Note to extension developers: The given uri is available
		to the action factory as "uri" attribute of the context.
		
		@param uri The URI (null is not allowed)
		@return An UriHandlerFactory or null if no factory was found
	*/
	public static ConnectionManagerFactory findFactory(URI uri)
	{
		if (uri == null) throw new NullPointerException("uri");
		
		ClientContext context = new ClientContext();
		context.setAttribute("uri", uri);
		
		ExtensionsInvoker<ConnectionManagerFactory> invoker = getInvoker(context, null);
		ConnectionManagerFactory factory = autoDetect(invoker, uri);
		return factory;
	}
	
	/**
		Utility method for creating a connection manager. If a specific factory
		is specified in the connection parameters, that factory is used, otherwise
		{@link #findFactory(URI)} is used for auto-detection. If no factory is found
		an exception is thrown, otherwise a connection manger is created by the
		factory method:
		{@link ConnectionManagerFactory#createConnectionManager(URI, ConnectionParameters)}
		
		@param uri The URI (required)
		@param parameters Connection parameters, such as user login/password, etc. (optional)
		@return A connection manager
	*/
	public static ConnectionManager createConnectionManager(URI uri, ConnectionParameters parameters)
	{
		if (uri == null) throw new NullPointerException("uri");
		ConnectionManagerFactory factory = null;
		if (parameters != null)
		{	
			uri = parameters.changeHost(uri);
			if (parameters.getConnectionManagerFactoryId() != null)
			{
				// Use specific factory
				String factoryId = parameters.getConnectionManagerFactoryId();
				factory = getFactory(factoryId);
				if (factory == null) 
				{
					throw new ItemNotFoundException("Connection manager factory: " + factoryId);
				}
			}
		}
		else
		{
			// Auto-detect factory base on URI
			factory = findFactory(uri);
			if (factory == null) 
			{
				throw new ItemNotFoundException("Connection manager factory for URI: " + uri);
			}
		}
		return factory.createConnectionManager(uri, parameters);
	}
	
	/**
		Utility method for creating a connection manager for a file item. If a 
		specific factory is specified in the connection parameters, that factory is used, 
		otherwise auto-detection is used. The auto-detection is similar
		to the {@link #findFactory(URI)} method but this time the context
		information for the action factory contains the file object and
		connection parameters.
		
		@param file The (external) file (required)
		@param parameters Connection parameters, such as user login/password, etc. (optional)
		@return A connection manager
	*/
	public static ConnectionManager createConnectionManager(File file, ConnectionParameters parameters)
	{
		if (file == null) throw new NullPointerException("file");
		// Create ClientContext for extension lookup
		URI uri = file.getURI();
		DbControl dc = file.getDbControl();
		ClientContext context = new ClientContext(dc, file);
		context.setAttribute("server", parameters);
		context.setAttribute("uri", uri);
		
		// Use a filter if a specific factory should be used
		ExtensionsFilter filter = null;
		if (parameters != null)
		{
			uri = parameters.changeHost(uri);
			if (parameters.getConnectionManagerFactoryId() != null)
			{
				filter = new SingleExtensionFilter(parameters.getConnectionManagerFactoryId());
			}
		}
		
		// Find extension
		ExtensionsInvoker<ConnectionManagerFactory> invoker = getInvoker(context, filter);
		ConnectionManagerFactory factory = filter == null ? autoDetect(invoker, uri) : getFirst(invoker);
		if (factory == null) 
		{
			throw new ItemNotFoundException("Connection manager for file: " + uri);
		}
		return factory.createConnectionManager(uri, parameters);
	}
	
	/**
		Get an invoker for using the connection manager factory extensions.
	*/
	@SuppressWarnings("unchecked")
	private static ExtensionsInvoker<ConnectionManagerFactory> getInvoker(ClientContext context, ExtensionsFilter filter)
	{
		ExtensionsManager manager = Application.getExtensionsManager();
		if (filter == null) filter = manager.getSettings();
		Registry registry = manager.getRegistry();
		return (ExtensionsInvoker<ConnectionManagerFactory>)registry.useExtensions(context, filter, EXTENSION_POINT_ID);
	}
	
	/**
		Get the first factory returned by the invoker.
	*/
	private static ConnectionManagerFactory getFirst(ExtensionsInvoker<ConnectionManagerFactory> invoker)
	{
		ConnectionManagerFactory factory = null;
		Iterator<ConnectionManagerFactory> it = invoker.iterator();
		if (it.hasNext()) factory = it.next();
		return factory;
	}

	/**
		Use auto-detection and return the first factory that can handle the given uri.
	*/
	private static ConnectionManagerFactory autoDetect(ExtensionsInvoker<ConnectionManagerFactory> invoker, URI uri)
	{
		for (ConnectionManagerFactory factory : invoker)
		{
			if (factory.useAutoDetection() && factory.supports(uri)) 
			{
				return factory;
			}
		}
		return null;
	}

}
