/* $Id: table.js 6576 2014-10-22 16:41:55Z nicklas $
	------------------------------------------------------------------
	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson, Gregory Vincic
	Copyright (C) 2007 Johan Enell, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
	------------------------------------------------------------------

	JavaScript functions for the Table taglib.

	@author Nicklas
	@version 2.0
*/
var Table = function()
{
	// Hold parsed column definitions -- initialized lazily when Table.getColumnDefs is called
	var columnDefs = [];

	var table = {};
	var internal = {};
	
	/**
		Initialize all control elements in the table, such as
		sorting, filtering, pageing, etc.
	*/
	internal.initTable = function(element, autoInit)
	{
		if (autoInit != 'table') return;
		var tableId = element.id;
		var tableFrm = document.forms[tableId];
		var itemType = Data.get(element, 'item-type');
		var attributes = {'table-id': tableId, 'item-type': itemType};
		
		// Add 'click' handler to page navigator
		// A single handler is added to the main div and relies 
		// on event bubbling from the actual targets
		var navigator = element.getElementsByClassName('table-navigator');
		for (var i = 0; i < navigator.length; i++)
		{
			Events.addEventHandler(navigator[i], 'click', Table.navigatorOnClick);
		}
		
		// Limit the 'rowsperpage' field to numbers and submit the form if ENTER is pressed
		var rowsPerPage = tableFrm.rowsperpage;
		if (rowsPerPage)
		{
			Events.addEventHandler(rowsPerPage, 'keypress', Events.integerOnly);
			Events.doOnEnter(rowsPerPage, Forms.submit);
		}
		
		// Add event handler for the 'view/preset' selection list
		// and icons for clearing the filter and show hidden columns with a filter
		if (tableFrm.presetselector)
		{
			Events.addEventHandler(tableFrm.presetselector, 'change', table.presetOnChange);
			Events.addEventHandler(tableId+'.clearfilter', 'click', table.clearFilterOnClick, attributes);
			Events.addEventHandler(tableId+'.showcolumns', 'click', table.showColumnsOnClick, attributes);
		}
		
		// Add 'click' handlers to column headers for sorting
		var sort = element.getElementsByClassName('table-sort');
		for (var i = 0; i < sort.length; i++)
		{
			Events.addEventHandler(sort[i], 'click', table.sortOnClick, attributes);
		}
		
		// Submit form when clicked or ENTER is pressed 
		var submit = element.getElementsByClassName('table-submit');
		for (var i = 0; i < submit.length; i++)
		{
			var field = submit[i];
			var type = field.type;
			if (type == 'radio' || type == 'checkbox')
			{
				// Radio and checkboxes react to 'click'
				Events.addEventHandler(field, 'click', Forms.submit);
			}
			else if (type == 'text')
			{
				// Text fields react to ENTER key
				Events.doOnEnter(field, Forms.submit);
			}
			else if (type == 'select-one')
			{
				// Selection list react to 'onchange'
				Events.addEventHandler(field, 'change', Forms.submit);
			}
		}
		
		// Add 'click' handler to the 'Check all' icon
		Buttons.addClickHandler('check.uncheck', table.checkUncheckOnClick, attributes);
		
		// Add 'click' handlers to all main table items
		// This must be different from the regular item click handler
		// since we must also support returning selected items from a
		// popup dialog (the regular click handler only supports navigating
		// to the 'view' or 'edit' page.
		var items = element.getElementsByClassName('table-item');
		for (var i = 0; i < items.length; i++)
		{
			var item = items[i];
			Buttons.addClickHandler(item, table.mainItemOnClick, attributes);
		}
		
		// Add 'click' handler to all 'Marked for removal' trashcan icons
		// This will delete the clicked item permanently
		var items = element.getElementsByClassName('table-delete-item');
		for (var i = 0; i < items.length; i++)
		{
			var item = items[i];
			Buttons.addClickHandler(item, Buttons.deleteItemPermanently, attributes);
		}

		// Add 'click' handler to all 'Shared item' icons
		// This will open the Share popup dialog 
		var items = element.getElementsByClassName('table-share-item');
		for (var i = 0; i < items.length; i++)
		{
			var item = items[i];
			Buttons.addClickHandler(item, Buttons.shareItem, attributes);
		}

	}
	Doc.addElementInitializer(internal.initTable);
	
	/**
		Click handler that changes the sort order of the table. The sort property
		and sort direction should be stored in the data-sort-property and data-sort-direction
		properties of the target element.
	*/
	table.sortOnClick = function(event)
	{
		var column = event.currentTarget;
		var sortProperty = Data.get(column, 'sort-property');
		var sortDirection = Data.get(column, 'sort-direction');
		var tableId = Data.get(column, 'table-id');
		var frm = document.forms[tableId];
		var mainDirection = frm.direction.value;
		if (frm.sortby.value != sortProperty || mainDirection != sortDirection)
		{
			if (event.altKey || event.ctrlKey || event.shiftKey) 
			{
				// Append the current sort property to the list of already sorted columns
				// If the sort property is already used we keep it, but use the new sort direction
				var sortArray = frm.sortby.value.split(',');
				var index = sortArray.indexOf(sortProperty);
				// Check for '+' and '-' prefix if the sort order is different from main direction
				if (index == -1) index = sortArray.indexOf('+'+sortProperty);
				if (index == -1) index = sortArray.indexOf('-'+sortProperty);
				// Currently not sorted, insert at end
				if (index == -1) index = sortArray.length;
				// Add the sort property as the last property
				if (sortDirection != mainDirection)
				{
					sortProperty = (sortDirection == 'ASC' ? '+' : '-') + sortProperty;
					sortDirection = mainDirection;
				}
				sortArray[index] = sortProperty;
				sortProperty = sortArray.join(',');
			}
			frm.sortby.value = sortProperty;
			frm.direction.value = sortDirection;
			frm.submit();
		}
	}
	
	/**
		Event handler for the page navigator. The actual target should 
		have a table-page attribute that specify the page to move to.
	*/
	table.navigatorOnClick = function(event)
	{
		var target = event.target;
		var page = Data.get(target, 'table-page');
		if (page == null) return;
		
		var tableId = Data.get(event.currentTarget, 'table-id');
		table.setPage(tableId, page);
	}
	
	/**
		Event handler for showing all hidden columns that have a filter.
		The hidden columns should be stored as a comma-separated list
		in 'data-hidden-columns' attribute and the element must have the
		table id stored in 'data-table-id' attribute.
	*/
	table.showColumnsOnClick = function(event)
	{
		var target = event.currentTarget;
		var hiddenColumns = Data.get(target, 'hidden-columns');
		var tableId = Data.get(target, 'table-id');
		table.showColumns(tableId, hiddenColumns);
	}
	
	/**
		Event handler for removing all filters from the table.
		The element must have the table id stored in 'data-table-id' 
		attribute.
	*/
	table.clearFilterOnClick = function(event)
	{
		var target = event.currentTarget;
		var tableId = Data.get(target, 'table-id');
		table.clearFilter(tableId);
	}
	
	/**
		Event handler for the 'view/presets' selection list.
	*/
	table.presetOnChange = function(event)
	{
		var presetSelector = event.currentTarget;
		var selected = presetSelector[presetSelector.selectedIndex];
		var cmd = selected.value;
		var tableId = presetSelector.form.name;
		
		// Reset the list to no selection
		presetSelector.selectedIndex = 0;		
		
		// Perform whatever command as indicated by the selected value
		if (cmd == 'save-as')
		{
			table.saveCurrentContext(tableId);
		}
		else if (cmd == 'manage')
		{
			table.manageContexts(tableId);
		}
		else if (cmd == 'load-preset')
		{
			var presetId = Data.get(selected, 'preset-id');
			table.loadContext(tableId, presetId);
		}
		else if (cmd == 'manage-columns')
		{
			table.configureColumns(tableId);
		}
		else if (cmd == 'clear-filter')
		{
			table.clearFilter(tableId);
		}
		else if (cmd == 'option')
		{
			var key = Data.get(selected, 'option-key');
			var value = Data.get(selected, 'option-value');
			table.setOption(tableId, key, value);
		}
	}

	/**
		Event handler for the 'Check/uncheck all' icon that select
		or deselect all checkboxes on the current table page.
		The element must have the table id stored in 'data-table-id' 
		attribute. The 'data-regexp' may be used to specify a regular
		expression to match checkbox names, otherwise all numeric 
		checkboxes are included.
	*/
	table.checkUncheckOnClick = function(event)
	{
		var target = event.currentTarget;
		var tableId = Data.get(target, 'table-id');
		var regexp = Data.get(target, 'regexp');
		Forms.checkUncheck(document.forms[tableId], regexp);
	}
	
	/**
		Event handler for clicking on the main item on the table. In normal
		list mode, the event handler simply forwards to the Item.itemOnClick
		method. In selection mode, the corresponding checkbox/radio button
		is selected and returned to the caller and the window is closed.
	*/
	table.mainItemOnClick = function(event)
	{
		var target = event.currentTarget;
		var tableId = Data.get(target, 'table-id');
		var itemId = Data.int(target, 'item-id');
		
		var frm = document.forms[tableId];
		var mode = frm.mode ? frm.mode.value : 'default';
		if (mode == 'selectone')
		{
			var index = Forms.checkRadio(frm.item_id, itemId);
			table.returnSelectedItems(tableId);
			App.closeWindow();
		}
		else if (mode == 'selectmultiple' || mode == 'selectmultiplenobuttons')
		{
			var checkbox = table.toggleCheckBox(tableId, itemId);
			table.returnSelectedItems(tableId);
			App.closeWindow();
		}
		else
		{
			Items.itemOnClick(event);
		}
	}
	
	/**
		Get all columns defined for the specified table.
		@param tableDiv The ID or main table element
	*/
	table.getColumnDefs = function(tableDiv)
	{
		tableDiv = Doc.element(tableDiv);
		var columns = columnDefs[tableDiv.id];
		if (!columns)
		{
			columns = Data.json(tableDiv.id+'.table-data', 'column-defs');
			// Store each definitition using the 'id' as a key
			for (var i = 0; i < columns.length; i++)
			{
				columns['id'+columns[i].id] = columns[i];
			}
			
			columnDefs[tableDiv.id] = columns;
		}
		return columns;
	}

	/**
		Change the page of the table.
		@param tableDiv The ID or main table element
		@param page The new page number
	*/
	table.setPage = function(tableDiv, page)
	{
		tableDiv = Doc.element(tableDiv);
		var frm = document.forms[tableDiv.id];
		if (frm.page.value != page)
		{
			frm.page.value = page;
			frm.submit();
		}
	}
	
	/**
		Change the visible columns of the table.
		@param tableDiv The ID or main table element
		@param columns Comma-separated list of column ID:s of the visible columns, or the value 'all'
	*/
	table.setColumns = function(tableDiv, columns)
	{
		tableDiv = Doc.element(tableDiv);
		var frm = document.forms[tableDiv.id];
		if (frm.columns.value != columns)
		{
			frm.columns.value = columns;
			frm.submit();
		}
	}

	/**
		Show columns in the table, preserving the currently visible columns.
		@param tableDiv The ID or main table element
		@param columns Comma-separated list of column ID:s of the visible columns, or the value 'all'
	*/
	table.showColumns = function(tableDiv, columns)
	{
		tableDiv = Doc.element(tableDiv);
		var frm = document.forms[tableDiv.id];
		if (frm.columns.value != columns)
		{
			frm.columns.value += ',' + columns;
			frm.submit();
		}
	}
	

	/**
		Get the visible columns of the table.
		@param tableDiv The ID or main table element
	*/
	table.getColumns = function(tableDiv)
	{
		tableDiv = Doc.element(tableDiv);
		var frm = document.forms[tableDiv.id];
		return frm.columns.value;
	}
	
	
	/**
		Clear the all filters 
		@param tableDiv The ID or main table element
	*/
	table.clearFilter = function(tableDiv)
	{
		tableDiv = Doc.element(tableDiv);
		var frm = document.forms[tableDiv.id];
		Forms.addHidden(frm, 'filter:clearAll', '1');
		frm.submit();
	}
	
	/**
		Open a popup window that allows a user to configure which columns
		that should be visible and their order.
		@param table The ID or main table element
		@param settingName The name of the setting that is used to store the
			column visibility and order, if not specified the default is 'columns'
	*/
	table.configureColumns = function(tableDiv, settingName)
	{
		tableDiv = Doc.element(tableDiv);
		var itemType = Data.get(tableDiv, 'item-type');
		var subContext = Data.get(tableDiv, 'subcontext', '');
		
		var url = App.getRoot()+'common/columns/configure.jsp?ID='+App.getSessionId();
		url += '&table_id='+tableDiv.id+'&item_type='+itemType+'&subcontext='+subContext;
		if (settingName) url += '&settingName='+settingName;
		Dialogs.openPopup(url, 'ConfigureColumns', 750, 450);
	}

	/**
		Open a popup dialog that allows the user to save the current 
		table settings to a named preset.
		@param table The ID or main table element
	*/
	table.saveCurrentContext = function(tableDiv)
	{
		tableDiv = Doc.element(tableDiv);
		var itemType = Data.get(tableDiv, 'item-type');
		var subContext = Data.get(tableDiv, 'subcontext', '');
		var url = App.getRoot()+'common/context/saveas.jsp?ID='+App.getSessionId();
		url += '&item_type='+itemType+'&subcontext='+subContext;
		Dialogs.openPopup(url, 'SaveCurrentContext', 450, 300);
	}
	
	/**
		Open a popup dialog that allows the user to manage (eg. load or delete)
		all table setting presets that have been saved for the given table.
		@param table The ID or main table element
	*/
	table.manageContexts = function(tableDiv)
	{
		tableDiv = Doc.element(tableDiv);
		var itemType = Data.get(tableDiv, 'item-type');
		var subContext = Data.get(tableDiv, 'subcontext', '');
		var url = App.getRoot()+'common/context/manage.jsp?ID='+App.getSessionId();
		url += '&item_type='+itemType+'&subcontext='+subContext;
		Dialogs.openPopup(url, 'ManageContexts', 450, 300);
	}
	
	/**
		Load the table settings preset with the given id.
		@param table The ID or main table element
	*/
	table.loadContext = function(tableDiv, contextId)
	{
		tableDiv = Doc.element(tableDiv);
		var frm = document.forms[tableDiv.id];
		Forms.addHidden(frm, 'context', contextId);
		frm.cmd.value = 'LoadContext';
		frm.submit();
	}
	
	/**
		Change a table option and reload the table.
		
		@param table The ID or main table element
		@param name The name of the option
		@param value The value of the option
	*/
	table.setOption = function(tableDiv, name, value)
	{
		tableDiv = Doc.element(tableDiv);
		var frm = document.forms[tableDiv.id];
		frm[name].value = value;
		frm.submit();
	}

	
	/**
		Open a popup window with the given width and height
		and submit the table form (POST) to that popup.
		@param tableDiv The ID or main table element
		@param cmd The command value to use when sending the form
			(the old command is restored afterwards)
		@param width The width in pixels of the popup (may be scaled)
		@param height The height in pixels of the popup (may be scaled)
	*/
	table.submitToPopup = function(tableDiv, cmd, width, height)
	{
		tableDiv = Doc.element(tableDiv);
		var frm = document.forms[tableDiv.id];
		var oldCmd = frm.cmd.value;
		var oldTarget = frm.target;
		frm.cmd.value = cmd;
		frm.target = tableDiv.id+cmd;
		Dialogs.openPopup('', frm.target, width, height);
		frm.submit();
		frm.target = oldTarget;
		frm.cmd.value = oldCmd;
	}
	
	/**
		Submit the table using the given cmd if there is 
		at least one item selected.
		@param tableDiv The ID or main table element
		@param cmd The command value to use when sending the form
		@param regexp A regular expression to use for matching checkboxes in the table,
			if not specified all checkboxes with a numeric name are used
	 */
	table.submitSelected = function(tableDiv, cmd, regexp)
	{
		tableDiv = Doc.element(tableDiv);
		var frm = document.forms[tableDiv.id];
		if (!table.checkIfSelected(tableDiv, regexp))
		{
			return;
		}
		frm.cmd.value = cmd;
		frm.submit();
	}
	
	/**
		Submit a request to delete (or put in trashcan) all selected items
		in the table.
		@param tableDiv The ID or main table element
		@param regexp A regular expression to use for matching checkboxes in the table,
			if not specified all checkboxes with a numeric name are used
		@param confirmFirst If set, ask for confirmation before deleting the items
	*/
	table.deleteItems = function(tableDiv, regexp, confirmFirst)
	{
		tableDiv = Doc.element(tableDiv);
		var frm = document.forms[tableDiv.id];
		var numSelected = table.checkIfSelected(tableDiv, regexp);
		if (numSelected == 0) return;
		if (confirmFirst)
		{
			if (!confirm('You are about to delete '+numSelected+' items. This can\'t be undone. Continue?'))
			{
				return;
			}
		}
		frm.cmd.value = 'DeleteItems';
		frm.submit();
	}
	
	/**
		Submit a request to undelete (remove from trashcan) all selected items
		in the table.
		@param tableDiv The ID or main table element
		@param regexp A regular expression to use for matching checkboxes in the table,
			if not specified all checkboxes with a numeric name are used
	*/
	table.restoreItems = function(tableDiv, regexp)
	{
		tableDiv = Doc.element(tableDiv);
		var frm = document.forms[tableDiv.id];
		if (!table.checkIfSelected(tableDiv, regexp))
		{
			return;
		}
		frm.cmd.value = 'RestoreItems';
		frm.submit();
	}

	
	/**
		Open a popup dialog that allows a user to select a new owner for
		all items that are selected in the given table. 
		@param tableDiv The ID or main table element
		@param regexp An optional regular expression matching the names
			of checkboxes that can be selected, if not specified it matches
			checkboxes with numeric names (eg. the name is the id of the item
			in the table)
	 */
	table.setOwnerOfItems = function(tableDiv, regexp)
	{
		tableDiv = Doc.element(tableDiv);
		if (!table.checkIfSelected(tableDiv, regexp))
		{
			return;
		}
		table.submitToPopup(tableDiv, 'SetOwnerOfItems', 450, 300);
	} 

	/**
		Open a popup dialog that allows a user to specify sharing permissions
		for all items that are selected in the given table. 
		@param tableDiv The ID or main table element
		@param regexp An optional regular expression matching the names
			of checkboxes that can be selected, if not specified it matches
			checkboxes with numeric names (eg. the name is the id of the item
			in the table)
	 */
	table.shareItems = function(tableDiv, regexp)
	{
		tableDiv = Doc.element(tableDiv);
		if (!table.checkIfSelected(tableDiv, regexp))
		{
			return;
		}
		table.submitToPopup(tableDiv, 'ShareItems', 600, 400);
	}

	/**
		Open a popup window for selecting a plug-in that can run in the 
		current list context
		@param tableDiv The ID or main table element
		@param pluginType One of: IMPORT, EXPORT, ANALYZE or OTHER (default)
		@param cmd Override the default command for running the plugin
	*/
	table.runPlugin = function(tableDiv, pluginType, cmd)
	{
		if (!cmd)
		{
			if (pluginType == 'EXPORT')
			{
				cmd = 'ExportItems';
			}
			else if (pluginType == 'IMPORT')
			{
				cmd = 'ImportItems';
			}
			else if (pluginType == 'ANALYZE')
			{
				cmd = 'RunListAnalysisPlugin';
			}
			else
			{
				cmd = 'RunListPlugin';
			}
		}
		table.submitToPopup(tableDiv, cmd, 750, 500);
	}

	/**
		Toggle the selected status of a checkbox with 
		@param tableDiv The ID or main table element
		
		@returns The toggled  element or null if no element was found
	*/
	table.toggleCheckBox = function(tableDiv, itemId)
	{
		tableDiv = Doc.element(tableDiv);
		var frm = document.forms[tableDiv.id];
		
		for (var i=0; i < frm.elements.length; i++)
		{
			var element = frm.elements[i];
			if (element.type == 'checkbox' && element.name == itemId)
			{
				element.checked = !element.checked;
				Events.sendChangeEvent(element);
				return element;
			}
		}
		return null;
	}

	/**
		Get the ID of all selected items in a table.
		
		@param tableDiv The ID or main table element
		@param regexp An optional regular expression matching the names
			of checkboxes that can be selected, if not specified it matches
			checkboxes with numeric names (eg. the name is the id of the item
			in the table)
		@return An array object with the ID:s
	*/
	table.getSelected = function(tableDiv, regexp)
	{
		tableDiv = Doc.element(tableDiv);
		if (!regexp) regexp = /\d+/;

		var frm = document.forms[tableDiv.id];
		var items = [];
		if (frm.item_id)
		{
			// Radio buttons
			var element = Forms.getCheckedRadio(frm.item_id);
			if (element != null)
			{
				items[items.length] = element.value;
			}
		}
		else
		{
			// Checkboxes
			for (var i=0; i < frm.elements.length; i++)
			{
				var element = frm.elements[i];
				if (element.type == 'checkbox' && element.name.match(regexp) && element.checked)
				{
					items[items.length] = element.value;
				}
			}
		}
		return items;
	}

	
	/**
		Return the selected items to the parent window using either a callback
		method (older way) or an event notification (new way).
		
		@param tableDiv The ID or main table element
		@param regexp An optional regular expression matching the names
			of checkboxes that can be selected, if not specified it matches
			checkboxes with numeric names (eg. the name is the id of the item
			in the table)

		@return The number of selected items
	*/
	table.returnSelectedItems = function(tableDiv, regexp)
	{
		tableDiv = Doc.element(tableDiv);
		if (!regexp) regexp = /\d+/;
		
		var frm = document.forms[tableDiv.id];
		var itemType = Data.get(tableDiv, 'item-type');
		
		// Should we use 'callback method' or 'event notification'?
		var callback = frm.callback ? frm.callback.value : null;
		var notifyTarget = callback ? window.top.opener.document.getElementById(callback) : null;
		var callbackMethod = callback ? window.top.opener[callback] : null;
		
		if (!notifyTarget && !callbackMethod) 
		{
			notifyTarget = tableDiv;
		}
		
		var selected = [];
		if (frm.item_id)
		{
			// Radio buttons allow single selection
			var element = Forms.getCheckedRadio(frm.item_id);
			if (element != null)
			{
				var id = element.value.match(/^\d+$/) ? parseInt(element.value) : element.value;
				selected[selected.length] = {'id': id, 'name': element.title, 'itemType': itemType};
			}
		}
		else
		{
			// Check boxes allow multiple selections
			for (var i=0; i < frm.elements.length; i++)
			{
				var element = frm.elements[i];
				if (element.type == 'checkbox' && element.name.match(regexp) && element.checked)
				{
					var id = element.value.match(/^\d+$/) ? parseInt(element.value) : element.value;
					selected[selected.length] = {'id': id, 'name': element.title, 'itemType': itemType};
				}
			}
		}
		
		var remaining = selected.length-1;
		for (var i = 0; i < selected.length; i++)
		{
			var s = selected[i];
			s.remaining = remaining--;
			if (notifyTarget)
			{
				// Send event to the target in the opener window
				Events.sendCustomEvent(notifyTarget, 'base-selected', s);
			}
			else if (callbackMethod)
			{
				// Call the callback method in the opener window
				callbackMethod.call(null, s.id, s.name);
			}
		}

		return selected.length;
	}

	table.checkIfSelected = function(tableDiv, regexp, messageIfNoneSelected)
	{
		tableDiv = Doc.element(tableDiv);
		var frm = document.forms[tableDiv.id];
		var numSelected = Forms.numChecked(frm, regexp);
		if (numSelected == 0)
		{
			if (!messageIfNoneSelected) 
			{
				messageIfNoneSelected = 'Please select at least one item in the list';
			}
			Forms.showNotification('check.uncheck', messageIfNoneSelected, null, 'pointer-left');
		}
		return numSelected;
	}

	return table;
}();


